"""Tests for the ACR Object Detectablity module."""

import pathlib
import unittest

from hazenlib.tasks.acr_object_detectability import ACRObjectDetectability
from hazenlib.utils import get_dicom_files

from tests import TEST_DATA_DIR, TEST_REPORT_DIR


class TestACRObjectDetectability(unittest.TestCase):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens")
    SCORE_8 = 10
    TOTAL_SCORE = 40

    def setUp(self):
        input_files = get_dicom_files(self.ACR_DATA)
        self.acr_object_detectability = ACRObjectDetectability(
            input_data=input_files,
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )
        self.results = self.acr_object_detectability.run()

    def test_slice_8_score(self):
        slice_8 = self.results.get_measurement(subtype="slice 8")[0].value
        assert slice_8 == self.SCORE_8

    def test_total_score(self):
        total_score = self.results.get_measurement(subtype="total")[0].value
        assert total_score == self.TOTAL_SCORE


class TestACRObjectDetectabilitySiemensSolaFit(TestACRObjectDetectability):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "SiemensSolaFit")
    SCORE_8 = 9
    TOTAL_SCORE = 38


class TestACRObjectDetectabilityGE(TestACRObjectDetectability):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE")
    SCORE_8 = 0  # in reality, this slice should be scored 0 but this is here to force future reassessments of the algorithm to check this dataset.
    TOTAL_SCORE = 0


class TestACRObjectDetectabilityPhilipsAchieva(TestACRObjectDetectability):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "PhilipsAchieva")
    SCORE_8 = 10
    TOTAL_SCORE = 40



"""Constants used throughout hazenlib."""

from typing import Literal

#######################################
# Allowed measurement types and names #
#######################################

MEASUREMENT_NAMES = Literal[
    "GeometricAccuracy",
    "Ghosting",
    "LowContrastObjectDetectability",
    "Relaxometry",
    "SlicePosition",
    "SliceWidth",
    "SNR",
    "SNRMap",
    "SpatialResolution",
    "Uniformity",
    "ExecutionMetadata",
]

MEASUREMENT_TYPES = Literal[
    "measured",
    "normalised",
    "fitted",
    "raw",
    "theoretical",
]

MEASUREMENT_VISIBILITY = Literal["final", "intermediate"]



import os
import unittest
import numpy as np

from hazenlib.ACRObject import ACRObject
from hazenlib.utils import dcmread
from tests import TEST_DATA_DIR


# Siemens (axial)
class TestACRTools(unittest.TestCase):
    centre = (129, 129)
    rotation = -1.0
    horizontal_distance = 191.40625
    horizontal_end = (129, 255)
    vertical_distance = 187.5
    vertical_end = (255, 128)

    def setUp(self):
        self.Siemens_data = [
            dcmread(os.path.join(TEST_DATA_DIR, "acr", "Siemens", f"{i}"))
            for i in os.listdir(os.path.join(TEST_DATA_DIR, "acr", "Siemens"))
        ]

        self.ACR_object = ACRObject(self.Siemens_data)
        self.img1 = self.ACR_object.slice_stack[0].pixel_array
        self.img7 = self.ACR_object.slice_stack[6].pixel_array

    def test_find_centre(self):
        phantom_centre, _ = self.ACR_object.find_phantom_center(
            self.img7, self.ACR_object.dx, self.ACR_object.dy
        )
        assert self.centre == phantom_centre

    def test_find_rotation(self):
        rotation_angle = self.ACR_object.determine_rotation(self.img1)
        assert self.rotation == np.round(rotation_angle, 1)

    def test_measure_orthogonal_lengths(self):
        cxy, _ = self.ACR_object.find_phantom_center(
            self.img1, self.ACR_object.dx, self.ACR_object.dy
        )
        mask = self.ACR_object.get_mask_image(self.img1, cxy)
        length_dict = self.ACR_object.measure_orthogonal_lengths(mask, cxy)
        # We report the distances with 2 significant figures of precision.
        assert np.round(self.horizontal_distance, 2) == np.round(
            length_dict["Horizontal Distance"], 2
        )
        assert self.horizontal_end == length_dict["Horizontal End"]
        assert np.round(self.vertical_distance, 2) == np.round(
            length_dict["Vertical Distance"], 2
        )
        assert self.vertical_end == length_dict["Vertical End"]


# GE axial
class TestACRToolsGE(TestACRTools):
    rotation = 0.0
    centre = (254, 255)
    horizontal_distance = 190.93
    horizontal_end = (256, 511)
    vertical_distance = 188.9016
    vertical_end = (511, 255)
    test_point = (-60.98, -45.62)

    def setUp(self):
        self.GE_data = [
            dcmread(os.path.join(TEST_DATA_DIR, "acr", "GE", f"{i}"))
            for i in os.listdir(os.path.join(TEST_DATA_DIR, "acr", "GE"))
        ]

        self.ACR_object = ACRObject(self.GE_data)
        self.img1 = self.ACR_object.slice_stack[0].pixel_array
        self.img7 = self.ACR_object.slice_stack[6].pixel_array

    def test_find_centre(self):
        phantom_centre, _ = self.ACR_object.find_phantom_center(
            self.img7, self.ACR_object.dx, self.ACR_object.dy
        )
        assert self.centre == phantom_centre

    def test_rotate_point(self):
        rotated_point = np.array(
            self.ACR_object.rotate_point((0, 0), (30, 70), 150)
        )
        rotated_point = np.round(rotated_point, 2)
        assert (rotated_point == self.test_point).all()


# Philips Achieva axial
class TestACRToolsPhilips(TestACRTools):
    rotation = 0.0
    centre = (129, 128)
    horizontal_distance = 190.4296875
    horizontal_end = (128, 255)
    vertical_distance = 189.453125
    vertical_end = (255, 129)

    def setUp(self):
        self.Philips_data = [
            dcmread(
                os.path.join(TEST_DATA_DIR, "acr", "PhilipsAchieva", f"{i}")
            )
            for i in os.listdir(
                os.path.join(TEST_DATA_DIR, "acr", "PhilipsAchieva")
            )
        ]

        self.ACR_object = ACRObject(self.Philips_data)
        self.img1 = self.ACR_object.slice_stack[0].pixel_array
        self.img7 = self.ACR_object.slice_stack[6].pixel_array

    def test_find_centre(self):
        phantom_centre, _ = self.ACR_object.find_phantom_center(
            self.img7, self.ACR_object.dx, self.ACR_object.dy
        )
        assert self.centre == phantom_centre


# Siemens Magnetom Sola Fit axial
class TestACRToolsSiemensSolaFit(TestACRTools):
    rotation = 0.0
    centre = (128, 127)
    horizontal_distance = 190.42959000000002
    horizontal_end = (127, 255)
    vertical_distance = 190.43
    vertical_end = (255, 128)

    def setUp(self):
        self.Philips_data = [
            dcmread(
                os.path.join(TEST_DATA_DIR, "acr", "SiemensSolaFit", f"{i}")
            )
            for i in os.listdir(
                os.path.join(TEST_DATA_DIR, "acr", "SiemensSolaFit")
            )
        ]

        self.ACR_object = ACRObject(self.Philips_data)
        self.img1 = self.ACR_object.slice_stack[0].pixel_array
        self.img7 = self.ACR_object.slice_stack[6].pixel_array

    def test_find_centre(self):
        phantom_centre, _ = self.ACR_object.find_phantom_center(
            self.img7, self.ACR_object.dx, self.ACR_object.dy
        )
        assert self.centre == phantom_centre



import unittest
import pathlib

import os
from tests import TEST_DATA_DIR, TEST_REPORT_DIR
from hazenlib.utils import get_dicom_files
from hazenlib.tasks.snr import SNR


# Note all SNR tests assume 5mm slice thickness
class TestSnr(unittest.TestCase):
    # SIEMENS MR_VE11C
    # 1.5T

    SNR_DATA = pathlib.Path(TEST_DATA_DIR / "snr")
    ORIENTATION = "Transverse"

    OBJECT_CENTRE = (
        131,
        122,
    )  # note these coordinates are (x, y) ie. (COLUMN, ROW)
    SNR_NORM_FACTOR = 9.761711312090041  # checked manually
    IMAGE_SMOOTHED_SNR = 1874.81  # this value from MATLAB for tra_250_2meas_1.IMA, single image smoothed, normalised
    IMAGE_SUBTRACT_SNR = 2130.93  # this value from MATLAB for tra_250_2meas_1.IMA and tra_250_2meas_2.IMA, subtract method, normalised

    # setting +/- 3% range for SNR results
    UPPER_SMOOTHED_SNR = IMAGE_SMOOTHED_SNR * 1.02
    LOWER_SMOOTHED_SNR = IMAGE_SMOOTHED_SNR * 0.90

    UPPER_SUBTRACT_SNR = IMAGE_SUBTRACT_SNR * 1.02
    LOWER_SUBTRACT_SNR = IMAGE_SUBTRACT_SNR * 0.98

    def setUp(self):
        self.snr = SNR(
            input_data=get_dicom_files(
                os.path.join(TEST_DATA_DIR, "snr", "Siemens"), sort=True
            ),
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )

    def test_get_object_centre(self):
        object_centre = self.snr.get_object_centre(self.snr.dcm_list[0])
        assert object_centre == self.OBJECT_CENTRE

    def test_image_snr(self):
        val = self.snr.run()
        img_desc = self.snr.img_desc(self.snr.dcm_list[0])
        smoothing_snr = val.get_measurement(
            name="SNR",
            subtype="smoothing",
            measurement_type="normalised",
            description=img_desc,
        )[0].value
        self.assertTrue(
            self.LOWER_SMOOTHED_SNR
            <= smoothing_snr
            <= self.UPPER_SMOOTHED_SNR,
        )
        subtract_snr = val.get_measurement(
            name="SNR",
            subtype="subtraction",
            measurement_type="normalised",
        )[0].value
        self.assertTrue(
            self.LOWER_SUBTRACT_SNR <= subtract_snr <= self.UPPER_SUBTRACT_SNR,
        )

    def test_SNR_factor(self):
        SNR_factor = self.snr.get_normalised_snr_factor(self.snr.dcm_list[0])
        assert (SNR_factor) == self.SNR_NORM_FACTOR


class TestSnrPhilips(TestSnr):
    # PHILIPS_MR_53_1
    # 1.5T

    SNR_DATA = pathlib.Path(TEST_DATA_DIR / "snr")
    ORIENTATION = "Coronal"

    OBJECT_CENTRE = (
        127,
        129,
    )  # note these coordinates are (x, y) ie. (COLUMN, ROW) taken from Hazen, but checked in close proximity to Matlab
    SNR_NORM_FACTOR = (
        14.35183536242098  # value taken from Hazen, but checked manually.
    )
    IMAGE_SMOOTHED_SNR = 5684.08  # this value from MATLAB for Philips_IM-0011-0005.dcm, single image smoothed, normalised
    IMAGE_SUBTRACT_SNR = 5472.44  # this value from MATLAB for Philips_IM-0011-0005.dcm and Philips_IM-0011-0006.dcm, subtract method, normalised

    # setting +/- 2% range for SNR results
    UPPER_SMOOTHED_SNR = IMAGE_SMOOTHED_SNR * 1.02
    LOWER_SMOOTHED_SNR = IMAGE_SMOOTHED_SNR * 0.98

    UPPER_SUBTRACT_SNR = IMAGE_SUBTRACT_SNR * 1.02
    LOWER_SUBTRACT_SNR = IMAGE_SUBTRACT_SNR * 0.98

    def setUp(self):
        # self.test_file = dcmread(str(self.SNR_DATA / 'Philips' / 'Philips_IM-0011-0005.dcm'), force=True)
        # self.test_file_2 = dcmread(str(self.SNR_DATA / 'Philips' / 'Philips_IM-0011-0006.dcm'), force=True)
        self.snr = SNR(
            input_data=get_dicom_files(
                os.path.join(TEST_DATA_DIR, "snr", "Philips"), sort=True
            ),
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )


class TestSnrGE(TestSnr):
    # GE ACCESSNET69-45B
    # 1.5T

    SNR_DATA = pathlib.Path(TEST_DATA_DIR / "snr")
    ORIENTATION = "Sagittal"

    OBJECT_CENTRE = (
        127,
        129,
    )  # note these coordinates are (x, y) ie. (COLUMN, ROW) taken from Hazen, but checked in close proximity to Matlab
    SNR_NORM_FACTOR = (
        8.254476647778304  # value taken from Hazen, but checked manually
    )
    IMAGE_SMOOTHED_SNR = 1551.19  # this value from MATLAB for GE_IM-0003-0001.dcm, single image smoothed, normalised
    IMAGE_SUBTRACT_SNR = 1517.88  # this value from MATLAB for GE_IM-0003-0001.dcm and Philips_IM-0004-0001.dcm, subtract method, normalised

    # setting +/- 2% range for SNR results
    UPPER_SMOOTHED_SNR = IMAGE_SMOOTHED_SNR * 1.02
    LOWER_SMOOTHED_SNR = IMAGE_SMOOTHED_SNR * 0.90

    UPPER_SUBTRACT_SNR = IMAGE_SUBTRACT_SNR * 1.02
    LOWER_SUBTRACT_SNR = IMAGE_SUBTRACT_SNR * 0.98

    def setUp(self):
        # self.test_file = dcmread(str(self.SNR_DATA / 'GE' / 'IM-0003-0001.dcm'), force=True)
        # self.test_file_2 = dcmread(str(self.SNR_DATA / 'GE' / 'IM-0004-0001.dcm'), force=True)
        self.snr = SNR(
            input_data=get_dicom_files(
                os.path.join(TEST_DATA_DIR, "snr", "GE"), sort=True
            ),
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )


class TestSnrThreshold(TestSnr):
    # SIEMENS VIDA
    # Example of shape detection failure

    SNR_DATA = pathlib.Path(TEST_DATA_DIR / "snr_threshold")
    SNR_DATA_2 = pathlib.Path(TEST_DATA_DIR / "snr")

    OBJECT_CENTRE = (129, 126)
    SNR_NORM_FACTOR = 13.537071812733949  # value taken from Hazen
    IMAGE_SMOOTHED_SNR = 5508.12  # TODO: get this value from Matlab, as in other tests (currently using Hazen value)
    IMAGE_SUBTRACT_SNR = 4809.91  # TODO: get this value from Matlab, as in other tests (currently using Hazen value)

    # setting +/- 2% range for SNR results
    UPPER_SMOOTHED_SNR = IMAGE_SMOOTHED_SNR * 1.02
    LOWER_SMOOTHED_SNR = IMAGE_SMOOTHED_SNR * 0.84

    UPPER_SUBTRACT_SNR = IMAGE_SUBTRACT_SNR * 1.02
    LOWER_SUBTRACT_SNR = IMAGE_SUBTRACT_SNR * 0.98

    def setUp(self):
        # self.test_file = dcmread(str(self.SNR_DATA / 'VIDA' / 'HC_SNR_SAG_1.dcm'), force=True)
        # self.test_file_2 = dcmread(str(self.SNR_DATA / 'VIDA' / 'HC_SNR_SAG_2.dcm'), force=True)
        self.snr = SNR(
            input_data=get_dicom_files(
                os.path.join(TEST_DATA_DIR, "snr_threshold", "VIDA"), sort=True
            ),
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )



"""A utility module for medical imaging analysis.

Providing:

- robust DICOM metadata extraction
- image processing
- ROI detection functions.

Supports multi-frame and vendor-specific DICOM formats
(GE, Siemens, Philips, Toshiba),
includes string scrubbing for security,
circular/shape detection via OpenCV,
parallel processing helpers, and debugging tools.

Designed for automated phantom QA in MRI workflows.

"""

from __future__ import annotations

# Python imports
import copy
import multiprocessing
import os
import re
from collections import defaultdict
from multiprocessing import Pool
from typing import Union, TypeVar

# Module imports
import cv2 as cv
import imutils
import matplotlib
import numpy as np
import pydicom
from skimage import filters, measure

# Local imports
import hazenlib.exceptions as exc
from hazenlib.logger import logger

matplotlib.use("Agg")

REGEX_SCRUBNAME = "\\^\\\\`\\{\\}\\[\\]\\(\\)\\!\\$'\\/\\ \\_\\:\\,\\-\\&\\=\\.\\*\\+\\;\\#"  #: Regex to match for these dirty characters.


def dcmread(*args, **kwargs) -> pydicom.dataset.FileDataset:
    """Thin wrapper around pydicom.dcmread.

    Provide default arguments and error handling.

    """
    try:
        return pydicom.dcmread(*args, **kwargs)

    except pydicom.errors.InvalidDicomError:
        logger.warning(
            "Couldn't read DICOM, trying with force=True"
            "\nargs:\n%s\n\nkwargs:\n%s",
            args,
            kwargs,
        )
        _ = kwargs.pop("force", True)
        return pydicom.dcmread(*args, force=True, **kwargs)


def scrub(dirtyString, matchCharacters, join_str="_"):
    """
    This function provides the core functionality for scrubbing strings for bad characters. This is the most useful
    function in the core library since it provides a security hardening benefit as well. Ideally, user input should get
    scrubbed with this function or derivatives. This functionality is mediated by `re.split
    <https://docs.python.org/3/library/re.html?highlight=re%20split#re.split>`_.

    :param dirtyString: Untrustworthy string that needs to be stripped of bad characters such as newlines.
    :param matchCharacters: Iterable of characters to scrub out of the string (typically a string of such characters)
    :param join_str: String used in-between clean fragments. Example, te\\nst => te_st if this parameter is _
    :return: Reconstructed clean string.
    """
    pattern = re.compile("[{}]".format(matchCharacters))
    target_string = re.split(pattern, str(dirtyString))
    return join_str.join(target_string)


def get_dicom_files(folder: str, sort=False) -> list:
    """Collect files with pixel_array into a list

    Args:
        folder (str): path to folder to check

    Returns:
        list: paths to DICOM image files (may be multi-framed)
    """
    file_list = []
    for file in os.listdir(folder):
        file_path = os.path.join(folder, file)
        if has_pixel_array(file_path):
            file_list.append(file_path)
    return file_list


def is_dicom_file(filename):
    """Check if file is a DICOM file, using the the first 128 bytes are preamble
    the next 4 bytes should contain DICM otherwise it is not a dicom

    Args:
        filename (str): path to file to be checked for the DICM header block

    Returns:
        bool: True or False whether file is a DICOM
    """
    # TODO: make it more robust, ensure that file contains a pixel_array
    file_stream = open(filename, "rb")
    file_stream.seek(128)
    data = file_stream.read(4)
    file_stream.close()
    if data == b"DICM":
        return True
    else:
        logger.debug("%s is not a DICOM file.", filename)
        return False


def has_pixel_array(filename) -> bool:
    """Check whether DICOM object has pixel_array that can be used for calc

    Args:
        filename (str): path to file to be checked

    Returns:
        bool: True/False whether pixel_array is available
    """

    try:
        dcm = dcmread(filename)
        # while enhanced DICOMs have a pixel_array, it's shape is in the format
        # (# frames, x_dim, y_dim)
        _ = dcm.pixel_array

    except AttributeError:
        logger.debug("%s does not contain image data", filename)
        return False

    else:
        return True


def is_enhanced_dicom(dcm: pydicom.Dataset) -> bool:
    """Check if file is an enhanced DICOM file

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Raises:
        Exception: Unrecognised_SOPClassUID

    Returns:
        bool: True or False whether file is an enhanced DICOM
    """

    if dcm.SOPClassUID in [
        "1.2.840.10008.5.1.4.1.1.4.1",
        "EnhancedMRImageStorage",
    ]:
        return True
    elif dcm.SOPClassUID == "1.2.840.10008.5.1.4.1.1.4":
        return False
    else:
        raise Exception("Unrecognised SOPClassUID")


def get_manufacturer(dcm: pydicom.Dataset) -> str:
    """Get the manufacturer field from the DICOM header

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Raises:
        Exception: _description_

    Returns:
        str: manufacturer of the scanner used to obtain the DICOM image
    """
    supported = ["ge", "siemens", "philips", "toshiba", "canon"]
    manufacturer = dcm.Manufacturer.lower()
    for item in supported:
        if item in manufacturer:
            return item

    msg = f"{manufacturer} not recognised manufacturer"
    logger.error(msg)
    raise Exception(msg)


def get_average(dcm: pydicom.Dataset) -> float:
    """Get the NumberOfAverages field from the DICOM header.

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Returns:
        float: value of the NumberOfAverages field from the DICOM header

    """
    if is_enhanced_dicom(dcm):
        averages = dcm[
            (0x5200, 0x9230)  # Per-Frame Functional Groups Sequence
        ][0][
            (0x0018, 0x9119)  # MR Averages Sequence
        ][0][
            (0x0018, 0x0083)  # Number of Averages
        ].value

    else:
        averages = dcm.NumberOfAverages

    return averages


def get_bandwidth(dcm: pydicom.Dataset) -> float:
    """Get the PixelBandwidth field from the DICOM header

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Returns:
        float: value of the PixelBandwidth field from the DICOM header
    """
    try:
        bandwidth = dcm.PixelBandwidth

    except AttributeError:
        bandwidth = dcm[
            (0x5200, 0x9229)  # Shared Functional Groups Sequence
        ][0][
            (0x0018, 0x9006)  # MR Imaging Modifier Sequence
        ][0][
            (0x0018, 0x0095)  # Pixel Bandwidth
        ].value
    return bandwidth


def get_num_of_frames(dcm: pydicom.Dataset) -> int:
    """Get the number of frames from the DICOM pixel_array

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Returns:
        float: value of the PixelBandwidth field from the DICOM header
    """
    # TODO: investigate what values could the dcm.pixel_array.shape be and what that means
    if len(dcm.pixel_array.shape) > 2:
        return dcm.pixel_array.shape[0]
    elif len(dcm.pixel_array.shape) == 2:
        return 1


def get_slice_thickness(dcm: pydicom.Dataset) -> float:
    """Get the SliceThickness field from the DICOM header

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Returns:
        float: value of the SliceThickness field from the DICOM header
    """
    if is_enhanced_dicom(dcm):
        try:
            slice_thickness = (
                dcm.PerFrameFunctionalGroupsSequence[0]
                .PixelMeasuresSequence[0]
                .SliceThickness
            )
        except AttributeError:
            slice_thickness = (
                dcm.PerFrameFunctionalGroupsSequence[0]
                .Private_2005_140f[0]
                .SliceThickness
            )
        except Exception:
            msg = "Unrecognised metadata Field for Slice Thickness"
            logger.exception(msg)
            raise Exception(msg)
    else:
        slice_thickness = dcm.SliceThickness

    return slice_thickness


def get_pixel_size(
    dcm: pydicom.Dataset,
    *,
    swap_indexes: bool = False,
) -> (float, float):
    """Get the PixelSpacing field from the DICOM header

    Args:
        dcm (pydicom.Dataset): DICOM image object
        swap_indexes : If True, will return (dy, dx) else (dx, dy)

    Returns:
        tuple of float: x and y values of the PixelSpacing field from the DICOM header

    """
    manufacturer = get_manufacturer(dcm)
    try:
        if is_enhanced_dicom(dcm):
            dx, dy = (
                dcm.PerFrameFunctionalGroupsSequence[0]
                .PixelMeasuresSequence[0]
                .PixelSpacing
            )
        else:
            dx, dy = dcm.PixelSpacing

    except AttributeError as err:
        logger.warning("Could not find PixelSpacing")
        if "ge" in manufacturer:
            fov = get_field_of_view(dcm)
            dx = fov / dcm.Columns
            dy = fov / dcm.Rows
        else:
            msg = "Manufacturer not recognised"
            logger.error(msg)
            raise ValueError(msg) from err

    if swap_indexes:
        return dy, dx
    return dx, dy


def get_TR(dcm: pydicom.Dataset, *, strict: bool = False) -> float:
    """Get the RepetitionTime field from the DICOM header.

    Args:
        dcm : DICOM image object

    Returns:
        return value : RepetitionTime field from the DICOM header.
            If not found and strict is False, defaults to 1000.

    Raises:
        AttributeError : If strict is True and no Reptition is found.

    """
    # TODO: explore what type of DICOM files do not have RepetitionTime in DICOM header
    try:
        if is_enhanced_dicom(dcm):
            TR = (
                dcm.SharedFunctionalGroupsSequence[0]
                .MRTimingAndRelatedParametersSequence[0]
                .RepetitionTime
            )
        else:
            TR = dcm.RepetitionTime

    except AttributeError as err:
        msg = "Could not find Repetition Time"
        if strict:
            raise AttributeError(msg) from err

        logger.warning("%s. Using default value of 1000 ms", msg)
        TR = 1000
    return TR


def get_rows(dcm: pydicom.Dataset) -> float:
    """Get the Rows field from the DICOM header.

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Returns:
        float: value of the Rows field from the DICOM header, or defaults to 256

    """
    try:
        rows = dcm.Rows

    except AttributeError:
        rows = 256
        logger.warning(
            "Could not find Number of matrix rows. Using default value of %i",
            rows,
        )

    return rows


def get_columns(dcm: pydicom.Dataset) -> float:
    """Get the Columns field from the DICOM header.

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Returns:
        float: value of the Columns field from the DICOM header, or defaults to 256

    """
    try:
        columns = dcm.Columns

    except AttributeError:
        columns = 256
        logger.warning(
            "Could not find matrix size (columns). Using default value of %i",
            columns,
        )
    return columns


def get_pe_direction(dcm: pydicom.Dataset):
    """Get the PhaseEncodingDirection field from the DICOM header.

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Returns:
        str: value of the InPlanePhaseEncodingDirection field from the DICOM header

    """
    if is_enhanced_dicom(dcm):
        return (
            dcm.SharedFunctionalGroupsSequence[0]
            .MRFOVGeometrySequence[0]
            .InPlanePhaseEncodingDirection
        )
    else:
        return dcm.InPlanePhaseEncodingDirection


def get_field_of_view(dcm: pydicom.Dataset):
    """Get Field of View value from DICOM header depending on manufacturer encoding

    Args:
        dcm (pydicom.Dataset): DICOM image object

    Raises:
        NotImplementedError: Manufacturer not GE, Siemens, Toshiba or Philips so FOV cannot be calculated.

    Returns:
        float: value of the Field of View (calculated as Columns * PixelSpacing[0])
    """
    # assumes square pixels
    manufacturer = get_manufacturer(dcm)

    if "ge" in manufacturer:
        fov = dcm[0x19, 0x101E].value
    elif "siemens" in manufacturer:
        fov = dcm.Columns * dcm.PixelSpacing[0]
    elif "philips" in manufacturer:
        if is_enhanced_dicom(dcm):
            fov = (
                dcm.Columns
                * dcm.PerFrameFunctionalGroupsSequence[0]
                .PixelMeasuresSequence[0]
                .PixelSpacing[0]
            )
        else:
            fov = dcm.Columns * dcm.PixelSpacing[0]
    elif "toshiba" in manufacturer:
        fov = dcm.Columns * dcm.PixelSpacing[0]
    else:
        logger.error("Manufacturer %s not supported", manufacturer)
        raise NotImplementedError(
            "Manufacturer not GE, Siemens, Toshiba or Philips so FOV cannot be calculated."
        )

    return fov


def get_datatype_max(dtype=np.uint8):
    """Get max value of the numpy datatype range.

    This is the machine max and not the data's max used value.
    For example, a np.uint8 has a machine range of 0 to 255.

    Args:
        dtype (np.dtype): Numpy datatype

    Returns:
        int|float: max machine value of data type

    """
    try:
        return np.iinfo(dtype).max
    except ValueError:
        return np.finfo(dtype).max


def get_datatype_min(dtype=np.uint8):
    """Get min value of the numpy datatype range.

    This is the machine min and not the data's min used value.
    For example, a np.uint8 has a machine range of 0 to 255.

    Args:
        dtype (np.dtype): Numpy datatype

    Returns:
        int|float: min machine value of data type

    """
    try:
        return np.iinfo(dtype).min
    except ValueError:
        return np.finfo(dtype).min


def get_image_IOP(dcm):
    """Get the IOP vectors from the DICOM header.

    Args:
        dcm (pydicom.dataset.FileDataset): DICOM Dataset

    Returns:
        pydicom.multival.MultiValue: vector
    """
    if is_enhanced_dicom(dcm):
        return (
            dcm.PerFrameFunctionalGroupsSequence[-1]
            .PlaneOrientationSequence[-1]
            .ImageOrientationPatient
        )
    return dcm.ImageOrientationPatient


def get_image_IPP(dcm):
    """Get the IPP vector from the DICOM header.

    Args:
        dcm (pydicom.dataset.FileDataset): DICOM Dataset

    Returns:
        pydicom.multival.MultiValue: vector
    """
    if is_enhanced_dicom(dcm):
        return (
            dcm.PerFrameFunctionalGroupsSequence[-1]
            .PlanePositionSequence[-1]
            .ImagePositionPatient
        )
    return dcm.ImagePositionPatient


def get_image_spacing(dcm):
    """Get the pixel resolution from the DICOM header.

    Args:
        dcm (pydicom.dataset.FileDataset): DICOM Dataset

    Returns:
        pydicom.multival.MultiValue: vector
    """
    if is_enhanced_dicom(dcm):
        return (
            dcm.PerFrameFunctionalGroupsSequence[-1]
            .PixelMeasuresSequence[-1]
            .PixelSpacing
        )
    return dcm.PixelSpacing


def get_image_orientation(dcm):
    """
    From http://dicomiseasy.blogspot.com/2013/06/getting-oriented-using-image-plane.html

    Args:
        dcm (list): values of dcm.ImageOrientationPatient - list of float

    Returns:
        str: Sagittal, Coronal or Transverse
    """
    iop = get_image_IOP(dcm)

    iop_round = [round(x) for x in iop]
    plane = np.cross(iop_round[0:3], iop_round[3:6])
    plane = [abs(x) for x in plane]
    if plane[0] == 1:
        return "Sagittal"
    elif plane[1] == 1:
        return "Coronal"
    elif plane[2] == 1:
        return "Transverse"


def determine_orientation(dcm_list):
    """Determine the phantom orientation based on DICOM metadata from a list of DICOM images.

    Note:
        The ImageOrientationPatient tag is a record of the orientation of the
        imaging volume which contains the phantom. The orientation of the
        imaging volume MAY NOT align with the true phantom orientation.

    Args:
        dcm_list (list): list of pyDICOM image objects.

    Returns:
        tuple (string, list):
            "saggital", "coronal", "axial", or "unexpected" orientation. \n
            list of the changing ImagePositionPatient values.
    """
    # Get the number of images in the list,
    # assuming each have a unique position in one of the 3 directions
    expected = len(dcm_list)

    iop = [np.round(c) for c in get_image_IOP(dcm_list[0])]
    x = np.array([round(get_image_IPP(dcm)[0]) for dcm in dcm_list])
    y = np.array([round(get_image_IPP(dcm)[1]) for dcm in dcm_list])
    z = np.array([round(get_image_IPP(dcm)[2]) for dcm in dcm_list])

    # Determine phantom orientation based on DICOM header metadata
    # Assume phantom orientation based on ImageOrientationPatient
    logger.debug(
        "Checking phantom orientation based on ImageOrientationPatient"
    )
    if iop == [0, 1, 0, 0, 0, -1] and len(set(x)) == expected:
        logger.debug("x %s", set(x))
        return "sagittal", x
    elif iop == [1, 0, 0, 0, 0, -1] and len(set(y)) == expected:
        logger.debug("y %s", set(y))
        return "coronal", y
    elif iop == [1, 0, 0, 0, 1, 0] and len(set(z)) == expected:
        logger.debug("z %s", set(z))
        return "axial", z
    else:
        logger.debug(
            "Checking phantom orientation based on ImagePositionPatient"
        )
        # Assume phantom orientation based on the changing value in ImagePositionPatient
        if (
            len(set(x)) == expected
            and len(set(y)) < expected
            and len(set(z)) < expected
        ):
            return "sagittal", x
        elif (
            len(set(x)) < expected
            and len(set(y)) == expected
            and len(set(z)) < expected
        ):
            return "coronal", y
        elif (
            len(set(x)) < expected
            and len(set(y)) < expected
            and len(set(z)) == expected
        ):
            return "axial", z
        else:
            logger.warning(
                "Unable to determine orientation based on DICOM metadata"
            )
            logger.info(f"Image orientation cosines => {iop}")
            logger.info("x %s\ny %s\nz %s", set(x), set(y), set(z))
            return "unexpected", [x, y, z]


def compute_dicom_frame_size(dcm):
    """Computes the full size in bytes of a Enhanced Multiframe DICOM frame. This value can be used for selecting the
    range of bytes for a given frame when manipulating the individual frames.

    Args:
        dcm (pydicom.dataset.FileDataset): DICOM Dataset

    Returns:
        float|int: byte count of a single frame
    """
    bits_allocated = dcm.BitsAllocated
    bytes_per_pixel = ((bits_allocated - 1) // 8) + 1
    samples_per_pixel = dcm.SamplesPerPixel
    rows, columns = dcm.Rows, dcm.Columns
    initial_frame_length = rows * columns * samples_per_pixel
    if bits_allocated == 1:
        return initial_frame_length // 8 + (
            initial_frame_length % 8 > 0
        )  # From pydicom's upcoming 3.0
    return initial_frame_length * bytes_per_pixel


def new_dicom(
    dcm: pydicom.dataset.FileDataset,
    pixel_data: np.ndarray,
    instance_number: int | None = None,
) -> pydicom.dataset.FileDataset:
    """Return a new DICOM with updated pixel data.

    Return a copy of dcm but with updated pixel data.
    Ensures Photometric Interpretation and Bits Stored
    are correct.

    Args:
        dcm : The original DICOM to copy.
        pixel_data : The updated pixel data.
        instance_number : If not None, used set the instance number.
            Defaults to None.

    Returns:
        new_dcm : A new dicom instance with the updated pixel data.

    """
    new_dcm = copy.deepcopy(dcm)

    # Extract single frame pixel data
    new_dcm.set_pixel_data(
        pixel_data,
        dcm[(0x0028, 0x0004)].value,  # Photometric Interpretation
        dcm[(0x0028, 0x0101)].value,  # Bits Stored
    )

    # Sets instance number
    new_dcm[(0x0020, 0x0013)].value = instance_number

    return new_dcm


def split_dicom(
    dcm: pydicom.dataset.FileDataset,
) -> list[pydicom.dataset.FileDataset]:
    """Split a multiframe DICOM into individual single-frame DICOMs.

    If the input is enhanced, we assume that it is a multiframe dicom
    and split it into constituent frames. We return this list.

    Otherwise, return a list whose single element is the given dicom

    Args:
        dcm : DICOM Dataset

    Returns:
        dcm_list : List of DICOM datasets

    """
    # In the case of non-enhanced DICOM - just wrap in a list
    if not is_enhanced_dicom(dcm):
        return [dcm]

    frame_count = dcm.NumberOfFrames
    single_frames = []

    for frame_idx in range(frame_count):
        try:
            pixel_data = dcm.pixel_array[frame_idx, :, :]

        # Case for (2D) sagittal localizer stored as an enhanced DICOM
        except IndexError:
            pixel_data = dcm.pixel_array

        single_frames.append(
            new_dicom(dcm, pixel_data, instance_number=frame_idx + 1),
        )

    return single_frames


def rescale_to_byte(array):
    """
    WARNING: This function normalises/equalises the histogram. This might have unintended consequences.

    Args:
        array (np.array): dcm.pixel_array

    Returns:
        np.array: normalised pixel values as 8-bit (byte) integer
    """
    image_histogram, bins = np.histogram(array.flatten(), 255)
    cdf = image_histogram.cumsum()  # cumulative distribution function
    cdf = 255 * cdf / cdf[-1]  # normalize

    # use linear interpolation of cdf to find new pixel values
    image_equalized = np.interp(array.flatten(), bins[:-1], cdf)

    return image_equalized.reshape(array.shape).astype("uint8")


def expand_data_range(
    data: np.array | np.ma.MaskedArray,
    valid_range: tuple | None = None,
    target_type: np.dtype = np.uint8,
) -> np.array | np.ma.MaskedArray:
    """Take a dataset and expands its data range to fill the value range possible in the target type.

    For example, if you have the data [1, 2, 3]
    and need it to fill the absolute values in uint16, you get:
    [0, 32767, 65,535]

    Args:
        data : dataset containing pixel values
        valid_range : tuple of values to use as the minimum and maximum
            of the dataset range.
            If None, we use the datasets' minimum and maximum.
        target_type : Numpy datatype to target in the expansion

    Returns:
        np.array: expanded range

    """
    dtype_max = get_datatype_max(target_type)
    lower, upper = (
        (data.min(), data.max()) if valid_range is None else valid_range
    )
    return (((data - lower) / (upper - lower)) * dtype_max).astype(target_type)


def detect_circle(img, dx):
    normalised_img = cv.normalize(
        src=img,
        dst=None,
        alpha=0,
        beta=255,
        norm_type=cv.NORM_MINMAX,
        dtype=cv.CV_8U,
    )
    detected_circles = cv.HoughCircles(
        normalised_img,
        cv.HOUGH_GRADIENT_ALT,
        1,
        param1=300,
        param2=0.9,
        minDist=int(10 / dx),
        minRadius=int(5 / (2 * dx)),
        maxRadius=int(16 / (2 * dx)),
    )
    if detected_circles is None:
        detected_circles = cv.HoughCircles(
            normalised_img,
            cv.HOUGH_GRADIENT,
            1,
            param1=50,
            param2=30,
            minDist=int(10 / dx),  # used to be 180 / dx
            minRadius=int(5 / dx),
            maxRadius=int(16 / dx),
        )
    # debug_image_sample(normalised_img)
    return detected_circles


def detect_centroid(img, dx, dy):
    """Attempt to detect circle locations using cv2.HoughCircles().

    We do the following preprocessing of the input to improve accuracy:

        #. Blur image with a boxcard kernel to help the Laplacian accentuate the circular features.
            It boosts the center accuracy by ~3% in the worst case I tested.
        #. Compute laplacian of the image to extract edges.
        #. Normalize to 8 bit.
        #. Attempt circle detection with HughesCircle Transform from OpenCV.

    These preprocessing steps improve the circle contours, which are then fed into the Hough Transform.
    By the way, we try the Hough Transform with cv.HOUGH_GRADIENT_ALT first. Failing that, we try the default Hough
    Transform mode. Also, we use 2 separate sets of parameter for each transform until we hopefully detect the phantom.

    Args:
        img (np.ndarray): pixel array containing the data to perform circle detection on
        dx (float): The coordinates of the point to rotate
        dy (float, optional): The amplitude threshold for peak identification. Defaults to 1.

    Returns:
        np.ndarray: Flattened array of tuples. tuples follow the form (x, y, r)

    """
    img_smoothed = cv.blur(img, (5, 5))
    img_grad = cv.Laplacian(img_smoothed, cv.CV_64F)
    img_grad_8u = cv.normalize(
        src=img_grad,
        dst=None,
        alpha=0,
        beta=255,
        norm_type=cv.NORM_MINMAX,
        dtype=cv.CV_8U,
    )

    try:
        detected_circles = cv.HoughCircles(
            img_grad_8u,
            cv.HOUGH_GRADIENT_ALT,
            1,
            param1=300,
            param2=0.9,
            minDist=int(180 / dy),
            minRadius=int(180 / (2 * dy)),
            maxRadius=int(200 / (2 * dx)),
        )
        if detected_circles is None:
            detected_circles = cv.HoughCircles(
                img_grad_8u,
                cv.HOUGH_GRADIENT,
                1,
                param1=50,
                param2=30,
                minDist=int(180 / dy),
                minRadius=int(180 / (2 * dy)),
                maxRadius=int(200 / (2 * dx)),
            )
    except AttributeError:
        detected_circles = cv.HoughCircles(
            img_grad_8u,
            cv.HOUGH_GRADIENT_ALT,
            1,
            param1=300,
            param2=0.9,
            minDist=int(180 / dy),
            minRadius=80,
            maxRadius=200,
        )
        if detected_circles is None:
            detected_circles = cv.HoughCircles(
                img_grad_8u,
                cv.HOUGH_GRADIENT,
                1,
                param1=50,
                param2=30,
                minDist=int(180 / dy),
                minRadius=80,
                maxRadius=200,
            )
    return detected_circles.flatten()


def compute_radius_from_area(area, voxel_resolution, conversion_value=10):
    """Calculates the radius of an ROI given an area. The radius is in pixel count. Meaning, if we want to get the
    radius for a 200cm2 ROI in a 0.5mm in-plane resolution, we call this function `with area = 200`, `voxel_resolution
    = 0.5`, and `conversion_value = 10`. This will yield a radius in mm which immediately gets divided by the
    resolution to yield the radius in pixel count units.

    Args:
        area (int): Area of ROI that will be generated with the radius calculated in this function
        voxel_resolution (float): voxel/pixel resolution as given by the PixelSpacing attribute in the DICOM header.
            This is typically in millimeter units.
        conversion_value (int): Value to use to convert the radius from area units (cm, etc) to mm.

    Returns:
        int: Integer radius length.
    """
    return np.ceil(
        np.divide(
            np.sqrt(np.divide(area, np.pi)) * conversion_value,
            voxel_resolution,
        )
    ).astype(int)


def create_cross_mask(img, length, x_coord, y_coord):
    """Generates a mask for an roi at the given coordinates

    Args:
        img (np.ndarray|np.ma.MaskedArray): pixel array containing the data where to generate roi
        radius (int): Integer radius of the circular roi
        x_coord (int): x coordinate of the center of the roi
        y_coord (int): y coordinate of the center of the roi

    Returns:
        np.ma.MaskedArray: Masked Array containing data for area of interest and zeros everywhere else.
    """
    grid = np.zeros(img.shape, dtype=np.bool_)

    height = int(length / 2)
    half_length = int(length / 2)
    half_height = int(height / 2)
    quarter_height = int(height / 4)

    x_start = int(x_coord - quarter_height)
    y_start = int(y_coord - half_length)
    grid[y_start : y_start + length, x_start : x_start + half_height] = True

    x_start = int(x_coord - half_length)
    y_start = int(y_coord - quarter_height)
    grid[y_start : y_start + half_height, x_start : x_start + length] = True
    return grid


def create_rectangular_mask(img, width, height, x_coord, y_coord):
    """Generates a mask for an roi at the given coordinates

    Args:
        img (np.ndarray|np.ma.MaskedArray): pixel array containing the data where to generate roi
        width (int): Integer radius of the circular roi
        height (int): Integer radius of the circular roi
        x_coord (int): x coordinate of the center of the roi
        y_coord (int): y coordinate of the center of the roi

    Returns:
        np.ma.MaskedArray: Masked Array containing data for area of interest and zeros everywhere else.
    """
    grid = np.zeros(img.shape, dtype=np.bool_)

    half_width = int(width / 2)
    half_height = int(height / 2)

    x_start = int(x_coord - half_width)
    y_start = int(y_coord - half_height)
    grid[y_start : y_start + height, x_start : x_start + width] = True
    return grid


def create_circular_mask(img, radius, x_coord, y_coord):
    """Generates a mask for an roi at the given coordinates

    Args:
        img (np.ndarray|np.ma.MaskedArray): pixel array containing the data where to generate roi
        radius (int): Integer radius of the circular roi
        x_coord (int): x coordinate of the center of the roi
        y_coord (int): y coordinate of the center of the roi

    Returns:
        np.ma.MaskedArray: Masked Array containing data for area of interest and zeros everywhere else.
    """
    height, width = img.shape
    y_grid, x_grid = np.ogrid[:height, :width]
    return (x_grid - x_coord) ** 2 + (y_grid - y_coord) ** 2 <= radius**2


def create_circular_kernel(radius):
    """Generate ROI kernel that can be used during convolutions. This is for generating circular kernels.

    Args:
        radius (int): Integer radius of the circular roi

    Returns:
        np.ndarray: Arrays of 1s and 0s comprising the circular kernel to use for convolution.
    """
    diameter = (radius * 2) + 1
    kernel_arr = np.zeros((diameter, diameter), dtype=np.bool_)
    return create_circular_mask(kernel_arr, radius, radius, radius).astype(
        np.int_
    )


def create_circular_mean_kernel(radius):
    """Generate ROI kernel that can be used during convolutions. This is for generating circular kernels.
    Uses :py:func:`create_roi_kernel` to generate the initial kernel mask.

    avg_kernel = kernel / kernel.sum()

    Convoluting against this kernel should yield

    Args:
        radius (int): Integer radius of the circular roi

    Returns:
        np.ndarray: Arrays of 1s and 0s comprising the circular kernel to use for convolution.
    """
    mask = create_circular_kernel(radius)
    return mask / mask.sum()


def create_cross_roi_at(img, width, x_coord, y_coord):
    """Generates a masked array delimiting the area of interest. It assists numpy in determining what data to use in
    math operations.

    Args:
        img (np.ndarray|np.ma.MaskedArray): pixel array containing the data where to generate roi
        radius (int): Integer radius of the circular roi
        x_coord (int): x coordinate of the center of the roi
        y_coord (int): y coordinate of the center of the roi

    Returns:
        np.ma.MaskedArray: Masked Array containing data for area of interest and zeros everywhere else.
    """
    mask = create_cross_mask(img, width, x_coord, y_coord)
    masked_img = np.ma.masked_array(img.copy(), mask=~mask, fill_value=0)
    return masked_img


def create_rectangular_roi_at(img, width, height, x_coord, y_coord):
    """Generates a masked array delimiting the area of interest. It assists numpy in determining what data to use in
    math operations.

    Args:
        img (np.ndarray|np.ma.MaskedArray): pixel array containing the data where to generate roi
        width (int): Integer radius of the circular roi
        height (int): Integer radius of the circular roi
        x_coord (int): x coordinate of the center of the roi
        y_coord (int): y coordinate of the center of the roi

    Returns:
        np.ma.MaskedArray: Masked Array containing data for area of interest and zeros everywhere else.
    """
    mask = create_rectangular_mask(img, width, height, x_coord, y_coord)
    masked_img = np.ma.masked_array(img.copy(), mask=~mask, fill_value=0)
    return masked_img


def create_circular_roi_at(img, radius, x_coord, y_coord):
    """Generates a masked array delimiting the area of interest. It assists numpy in determining what data to use in
    math operations.

    Args:
        img (np.ndarray|np.ma.MaskedArray): pixel array containing the data where to generate roi
        radius (int): Integer radius of the circular roi
        x_coord (int): x coordinate of the center of the roi
        y_coord (int): y coordinate of the center of the roi

    Returns:
        np.ma.MaskedArray: Masked Array containing data for area of interest and zeros everywhere else.
    """
    mask = create_circular_mask(img, radius, x_coord, y_coord)
    masked_img = np.ma.masked_array(img.copy(), mask=~mask, fill_value=0)
    return masked_img


def create_circular_roi_with_numpy_index(img, radius, argx):
    """Wrapper around :py:func:`create_roi_at` meant to use flat element indices and return an roi centered around
    this element.

    Args:
        img (np.ndarray): pixel array containing the data where to generate roi
        radius (int): Integer radius of the circular roi
        argx (int): index to nd.array element if the array was flattened. Example, value from argmax().

    Returns:
        np.ma.MaskedArray: Masked Array containing data for area of interest and zeros everywhere else.
    """
    x_coord, y_coord = detect_roi_center(img, argx)
    return (
        create_circular_roi_at(img, radius, x_coord, y_coord),
        x_coord,
        y_coord,
    )


def detect_roi_center(img, argx):
    """Finds the x and y coordinates of the center of an roi given the flat index in the numpy array!

    Args:
        img (np.ndarray): pixel array containing the data where to generate roi
        argx (int): index to nd.array element if the array was flattened. Example, value from argmax().

    Returns:
        x_coord (int): x coordinate of the center of the roi
        y_coord (int): y coordinate of the center of the roi
    """
    height, width = img.shape
    y, x = np.divmod(
        argx, width
    )  # returns x //y, x % y per docs but x is found with x % y
    return x, y


def wait_on_parallel_results(fxn, arg_list=None, *, debug: bool = False):
    """Parallelises a function into n number of jobs. It uses Python's multiprocessing Pool to spawn several processes
    that accept each job instance and processes it. The main use in this project is as a way to keep the report writing
    as fast as possible when we have multiple images to write to disk.

    Args:
        fxn (function): function symbol to execute on arguments
        arg_list (list of tuple): List of tuples. Each tuple has the list of parameters to pass to function. Therefore,
            each tuple symbolizes a job we need to process using the specified function.
        debug : If True, runs sequentially rather than using parallel.

    Returns:
        list: List of values returned by each job.
    """
    arg_list = () if arg_list is None else arg_list
    # Check if already in a daemon process
    if multiprocessing.current_process().daemon or debug:
        return [fxn(*args) for args in arg_list]

    with Pool() as pool:
        results = []
        result_handles = []
        for args in arg_list:
            result_handles.append(pool.apply_async(fxn, args))

        pool.close()
        pool.join()

        for r in result_handles:
            results.append(r.get())
        return results


def debug_image_sample(img, out_path=None):
    """Uses :py:class:`DebugSnapshotShow` to display the current image snapshot.
    Use this function to force a display of an intermediate numpy image array to visually inspect results.

    Args:
        img (np.ndarray): pixel array containing the data to display
        out_path (str): file path where you would like to save a copy of the image

    """
    if len(img):
        snapshot = DebugSnapshotShow(img).image
        if out_path is not None:
            snapshot.save(out_path, format="PNG", dpi=(300, 300))


def debug_plot_sample(img, plot_indx=0):
    """Uses :py:class:`DebugSnapshotShow` to display the current image snapshot.
    Use this function to force a display of an intermediate numpy image array to visually inspect results.

    Args:
        img (np.ndarray): pixel array containing the data to display
        out_path (str): file path where you would like to save a copy of the image

    """
    if len(img):
        import matplotlib.pyplot as plt

        plt.plot(img)
        plt.savefig(f"/tmp/hazen_debug_plot_{plot_indx}.png")


def debug_image_sample_circles(img, circles=[], out_path=None):
    """Uses :py:class:`DebugSnapshotShow` to display the current image snapshot.
    Use this function to force a display of an intermediate numpy image array to visually inspect results.

    Args:
        img (np.ndarray): pixel array containing the data to display
        out_path (str): file path where you would like to save a copy of the image

    """
    for circle in circles[-1]:
        logger.info(f"Center {circle[0]}, {circle[1]}")
        center = (int(circle[0]), int(circle[1]))
        cv.circle(img, center, int(circle[2]), (0, 255, 0), 1)
    debug_image_sample(img, out_path)


class DebugSnapshotShow:
    """
    This class manages presentation of an image (file path or instance of PIL.Image. This class is used as if it were
    a function.
    See the `Pillow ImageShow Documentation <https://pillow.readthedocs.io/en/stable/reference/ImageShow.html>`_.
    You will need to install Pillow/PIL library and dependencies separately.
    This class is meant to assist during debugging of image processing steps.
    """

    def __init__(self, image_instance):
        from PIL import Image, ImageShow

        if isinstance(image_instance, str):
            image_instance = Image.open(image_instance)
        elif isinstance(image_instance, np.ndarray) or isinstance(
            image_instance, np.ma.MaskedArray
        ):
            image_instance = expand_data_range(
                image_instance, target_type=np.uint8
            )
            image_instance = Image.fromarray(image_instance)
        presenter = ImageShow.EogViewer()
        presenter.show_image(image_instance)
        self.image = image_instance


class Rod:
    """Class for rods detected in the image"""

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        return f"Rod: {self.x}, {self.y}"

    def __str__(self):
        return f"Rod: {self.x}, {self.y}"

    @property
    def centroid(self):
        return self.x, self.y

    def __lt__(self, other):
        """Using "reading order" in a coordinate system where 0,0 is bottom left"""
        try:
            x0, y0 = self.centroid
            x1, y1 = other.centroid
            return (-y0, x0) < (-y1, x1)
        except AttributeError:
            logger.error(
                "Object other (type %s) has no property 'centroid'"
                " which is required for comparison.",
                type(other),
            )
            return NotImplemented

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y


class ShapeDetector:
    """Class for the detection of shapes in pixel arrays
    This class is largely adapted from https://www.pyimagesearch.com/2016/02/08/opencv-shape-detection/
    """

    def __init__(self, arr):
        self.arr = arr
        self.contours = None
        self.shapes = defaultdict(list)
        self.blurred = None
        self.thresh = None

    def find_contours(self):
        """Find contours in pixel array"""
        # convert the resized image to grayscale, blur it slightly, and threshold it
        self.blurred = cv.GaussianBlur(
            self.arr.copy(), (5, 5), 0
        )  # magic numbers

        optimal_threshold = filters.threshold_li(
            self.blurred, initial_guess=np.quantile(self.blurred, 0.50)
        )
        self.thresh = np.where(
            self.blurred > optimal_threshold, 255, 0
        ).astype(np.uint8)

        # have to convert type for find contours
        contours = cv.findContours(self.thresh, cv.RETR_TREE, 1)
        self.contours = imutils.grab_contours(contours)
        # rep = cv.drawContours(self.arr.copy(), [self.contours[0]], -1, color=(0, 255, 0), thickness=5)
        # plt.imshow(rep)
        # plt.title("rep")
        # plt.colorbar()
        # plt.show()

    def detect(self):
        """Detect specified shapes in pixel array

        Currently supported shapes:
            - circle
            - triangle
            - rectangle
            - pentagon
        """
        for c in self.contours:
            # initialize the shape name and approximate the contour
            peri = cv.arcLength(c, True)
            if peri < 100:
                # ignore small shapes, magic number is complete guess
                continue
            approx = cv.approxPolyDP(c, 0.04 * peri, True)

            # if the shape is a triangle, it will have 3 vertices
            if len(approx) == 3:
                shape = "triangle"

            # if the shape has 4 vertices, it is either a square or
            # a rectangle
            elif len(approx) == 4:
                shape = "rectangle"

            # if the shape is a pentagon, it will have 5 vertices
            elif len(approx) == 5:
                shape = "pentagon"

            # otherwise, we assume the shape is a circle
            else:
                logger.debug(
                    "%i vertices detected - assuming a circle", len(approx)
                )
                shape = "circle"

            # return the name of the shape
            logger.debug("%s detected", shape)
            self.shapes[shape].append(c)

    def get_shape(self, shape):
        """Identify shapes in pixel array

        Args:
            shape (_type_): _description_

        Raises:
            exc.ShapeDetectionError: ensure that only expected shapes are detected
            exc.MultipleShapesError: ensure that only 1 shape is detected

        Returns:
            tuple: varies depending on shape detected
                - circle: x, y, r - corresponding to x,y coords of centre and radius
                - rectangle/square: (x, y), size, angle - corresponding to x,y coords of centre, size (tuple) and angle in degrees
        """
        self.find_contours()
        self.detect()

        if shape not in self.shapes:
            logger.error(
                "No valid shape detected - got %s but expected one of %s",
                shape,
                self.shapes.keys(),
            )
            raise exc.ShapeDetectionError(shape)

        if len(self.shapes[shape]) > 1:
            shapes = [
                {shape: len(contours)}
                for shape, contours in self.shapes.items()
            ]
            logger.error(
                "Multiple (%i) shapes were detected - should be just 1",
                len(self.shapes[shape]),
            )
            raise exc.MultipleShapesError(shapes)

        contour = self.shapes[shape][0]
        if shape == "circle":
            # (x,y) is centre of circle, in x, y coordinates. x=column, y=row.
            (x, y), r = cv.minEnclosingCircle(contour)
            return x, y, r

        # Outputs in below code chosen to match cv.minAreaRect output
        # https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_imgproc/py_contours/py_contour_features/py_contour_features.html#b-rotated-rectangle
        # (x,y) is top-left of rectangle, in x, y coordinates. x=column, y=row.

        if shape == "rectangle" or shape == "square":
            (x, y), size, angle = cv.minAreaRect(contour)
            # OpenCV v4.5 adjustment
            # - cv.minAreaRect() output tuple order changed since v3.4
            # - swap size order & rotate angle by -90
            #
            # The above seems to have reverted in OpenCV v4.12
            # https://github.com/opencv/opencv/issues/28051
            # so the corrections are no longer needed.
            return (x, y), size, angle


P = TypeVar("P", bound="Point")
L = TypeVar("L", bound="Line")
xy = TypeVar("xy", bound="XY")


class XY(np.ndarray):
    """Class for 2D numpy array for plotting"""

    def __new__(cls, *args: list[Union[int, float]]):
        """Initialise class"""
        if len(set(map(len, args))) != 1:
            raise ValueError("All input arrays must have the same length")
        arr = np.array(args)
        if arr.ndim != 2 or arr.shape[0] != 2:
            raise ValueError("args of XY should be two 1d arrays")
        return np.asarray(arr).view(cls)

    @property
    def x(self) -> xy:
        """Property for x array of plotting series"""
        return self[0]

    @property
    def y(self) -> xy:
        """Property for y array of plotting series"""
        return self[1]

    @y.setter
    def y(self, val: np.ndarray):
        """Setter for y property"""
        if isinstance(val, (np.ndarray, list)):
            if len(val) != len(self.y):
                raise ValueError("Cannot modify shape of XY.y")
        else:
            raise TypeError(
                "Expected input to be either a list or numpy.ndarray"
            )
        self[1] = val


class Point(np.ndarray):
    """Class for 2D spatial point"""

    def __new__(cls, *args: Union[int, float]) -> np.ndarray:
        """Initialise the point class."""
        supported_dims = {2}
        if len(args) not in supported_dims:
            err = (
                f"Only {' or '.join(str(d) + 'D' for d in supported_dims)}"
                " points are supported"
            )
            raise NotImplementedError(err)

        return np.asarray(args, dtype=float).view(cls)

    @property
    def x(self) -> float:
        """Property for x coordinate"""
        return self[0]

    @property
    def y(self) -> float:
        """property for y coordinate"""
        return self[1]

    def get_distance_to(self, other: P) -> float:
        """Calculates distance between two point objects

        Args:
            other: Point to calculate distance to

        Returns:
            dist: The distance between the points

        """

        if not isinstance(other, Point):
            err = f"other should be Point and not {type(other)}"
            raise TypeError(err)
        return np.sqrt(np.sum((self - other) ** 2)).item()

    def __iter__(self):
        """Get iterable for plotting"""
        return iter([self.x, self.y])


class Line:
    """Class for line joining two points"""

    def __init__(self, *args: Point) -> None:
        """Initialise Line object"""

        for arg in args:
            if not isinstance(arg, Point):
                err = "All positional args should be instances of Point"
                raise TypeError(err)

        if len(args) != 2:
            err = "Only two positional args should be provided (start and end points)"
            raise ValueError(err)

        self.start = args[0]
        self.end = args[1]
        self.midpoint = (self.start + self.end) / 2

    def get_signal(self, refImg: np.ndarray) -> None:
        """Gets signal across line using pixel values from reference image

        Args:
            refImg (np.ndarray): Reference image for obtaining pixel values

        """
        signal = measure.profile_line(
            image=refImg,
            src=self.start[::-1].astype(int).tolist(),
            dst=self.end[::-1].astype(int).tolist(),
        )
        self.signal = XY(range(len(signal)), signal)

    def get_subline(self, perc: Union[int, float]) -> L:
        """Returns a "subline" of self.
        This is a line that shares the same unit vector but is reduced in length.
        Length of subline set to be "perc" percent of the length of self.

        Args:
            perc (int, float): controls length of subline. Subline will be "perc" percent of original length.

        Returns:
            subline (Line): subline of original line

        """
        if not isinstance(perc, (int, float)):
            err = f"perc should be int or float, not {type(perc)}."
            raise TypeError(err)

        if not 0 < perc <= 100:
            err = f"perc should be in bounds (0, 100] but received {perc}"
            raise ValueError(err)

        percOffSide = (100 - perc) / 2
        vector = self.start - self.end

        start = self.start - vector * percOffSide / 100
        end = self.end + vector * percOffSide / 100

        return type(self)(start, end)

    def point_swap(self):
        """Swaps start and end points and reverses associated attributes
        Args:
            None
        Returns:
            None
        """
        self.start, self.end = self.end, self.start
        if hasattr(self, "signal"):
            self.signal = self.signal[::-1]

    def __iter__(self) -> iter:
        """Get iterable for plotting"""
        points = (self.start, self.end)
        return iter([[p.x for p in points], [p.y for p in points]])

    def __str__(self):
        """Get string representation"""
        s = f"Line(start={self.start}, end={self.end})"
        return s



"""Types used for hazenlib."""

from __future__ import annotations

from pathlib import Path

# Typing imports
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator

    from numpy.typing import NDArray

# Python imports
import functools
import json
import logging
from collections.abc import Sequence
from dataclasses import asdict, dataclass, field, InitVar
from enum import Enum
from typing import Any, ParamSpec, get_args

# Module imports
import numpy as np
import pydicom
import scipy as sp

# Local imports
from hazenlib._version import __version__
from hazenlib.constants import (
    MEASUREMENT_NAMES,
    MEASUREMENT_TYPES,
    MEASUREMENT_VISIBILITY,
)
from hazenlib.exceptions import (
    InvalidMeasurementNameError,
    InvalidMeasurementTypeError,
    InvalidMeasurementVisibilityError,
)
from hazenlib.utils import get_pixel_size

logger = logging.getLogger(__name__)

#########################################
# ParamSpec for public methods/function #
#########################################

P_HazenTask = ParamSpec("P_HazenTask")

#############################
# Task Description Metadata #
#############################


class PhantomType(Enum):
    """Supported phantom types for QA tasks"""

    ACR = "ACR"
    MAGNET = "MagNET"
    CALIBER = "Caliber"


@dataclass
class TaskMetadata:
    module_name: str
    class_name: str
    single_image: bool = False
    phantom: PhantomType | None = None
    requires_args: list[str] | None = None


################
# Base Classes #
################


class JsonSerializableMixin:
    """Mix-in that supplies a shallow dict and json representation."""

    def to_dict(self) -> dict[str, Any]:
        """Return a shallow dictionary representation of the instance."""
        try:
            data = asdict(self)  # type: ignore[call-overload]

        except TypeError as err:
            logger.debug(
                "JsonSerializableMixin should only really be inhereted from by"
                " dataclasses but got  %s which raised error %s. "
                "Falling back to using __dict__",
                type(self),
                err,
            )
            data = dict(self.__dict__)
        return {k: v for k, v in data.items() if not k.startswith("_")}

    def to_json(
        self,
        *,
        indent: int | None = 2,
        sort_keys: bool = False,
        ensure_ascii: bool = True,
        **extra_kwargs: Any,  # noqa: ANN401
    ) -> str:
        """Serialize JSON with common formatting options.

        Args:
            indent : Number of spaces for pretty printing. None for compact output.
                     Defaults to 2.
            sort_keys : Alphabetically sort object keys. Default is False.
            ensure_ascii : Escape non-ASCII characters. Default is True.
            **extra_kwargs : Addition parameters passed directly to json.dumps().
                    See Python's json module documentation for available options.

        Notes:
            This method intentionally exposes only the most commonly used
            parameters with specific defaults. Advanced use cases can leverage
            extra_kwargs for parameters like `cls` (custom encoder) or `default`
            (fallback serializer).

        """
        payload = self.to_dict()
        return json.dumps(
            payload,
            indent=indent,
            sort_keys=sort_keys,
            ensure_ascii=ensure_ascii,
            **extra_kwargs,
        )


####################################################
# The canonical result that every task must return #
####################################################


@dataclass(frozen=True, slots=True)
class Measurement(JsonSerializableMixin):
    """Canonical result each measurment must have."""

    name: MEASUREMENT_NAMES
    value: float | NDArray[np.float64]
    type: MEASUREMENT_TYPES = "measured"
    subtype: str = ""
    description: str = ""
    unit: str = ""
    # For now we'll assume that we'll need to separate results
    # by final or intermediate. However, we'll define a variable
    # `visibility` as a string to leave the door open for future
    # implementations that may need more fine grained control.
    visibility: MEASUREMENT_VISIBILITY = "final"

    def __post_init__(self) -> None:
        """Validate the measurement inputs."""
        if self.name not in get_args(MEASUREMENT_NAMES):
            raise InvalidMeasurementNameError(self.name)

        if self.type not in get_args(MEASUREMENT_TYPES):
            raise InvalidMeasurementTypeError(self.type)

        if self.visibility not in get_args(MEASUREMENT_VISIBILITY):
            raise InvalidMeasurementVisibilityError(self.visibility)

    @property
    def final(self) -> bool:
        """Convenience accessor for filtering logic."""
        return self.visibility == "final"


@dataclass(slots=True)
class Metadata(JsonSerializableMixin):
    """Canonical dictionary for result metadata."""

    files: Sequence[str] | None = None
    institution_name: str | None = None
    manufacturer: str | None = None
    model: str | None = None
    date: str | None = None
    series_id: str | None = None
    study_id: str | None = None
    acquisition_number: int | str | None = None
    version: str | None = None

    def __post_init__(self) -> None:
        """Populate missing information if needed."""
        if self.version is None:
            self.version = __version__

        # If no files are present then no other information
        # can be gathered
        if self.files is None:
            return

        # This does involve likely reading the files twice but the overhead
        # of this is very minimal and results in minimum changes to existing
        # code base. If there is ever a significant refactor or this poses
        # a performance overhead then it's worth passing the DICOM objects
        # directly to results or metadata.
        try:
            dcm_list = [
                pydicom.dcmread(f, stop_before_pixels=True)
                for f in self.files
                if Path(f).suffix not in {".yml", ".yaml", ".toml", ".json"}
            ]

        except Exception as e:  # noqa: BLE001
            logger.warning("Failed to read DICOM files: %s", e)
            return

        if not dcm_list:
            return

        _dicom_tags = {
            "institution_name": "InstitutionName",
            "manufacturer": "Manufacturer",
            "model": "ManufacturerModelName",
            "date": "StudyDate",
            "series_id": "SeriesInstanceUID",
            "study_id": "StudyInstanceUID",
            "acquisition_number": "AcquisitionNumber",
        }

        for _field, tag in _dicom_tags.items():
            # Skip if already provided
            if getattr(self, _field) is not None:
                continue

            # Collect non-empty unique values
            values = {
                str(getattr(d, tag))
                for d in dcm_list
                if hasattr(d, tag) and getattr(d, tag)
            }

            if len(values) > 1:
                logger.warning(
                    "Multiple %s values detected: %s",
                    _field,
                    values,
                )

            if values:
                setattr(self, _field, next(iter(values)))


@dataclass
class Result(JsonSerializableMixin):
    """Canonical result any Task.run() must return."""

    task: str
    desc: str = ""
    files: str | Sequence[str] | Sequence[Sequence[str]] | None = None
    _load_metadata: InitVar[bool] = True

    def __post_init__(self, _load_metadata: bool) -> None:
        """Initialize the measurements, report_images and metadata."""
        self._measurements: list[Measurement] = []
        self._report_images: list[str] = []

        if _load_metadata:
            self.metadata = Metadata(files=self.files)

    @property
    def measurements(self) -> tuple[Measurement, ...]:
        """Tuple of result measurements."""
        return tuple(self._measurements)

    @property
    def report_images(self) -> tuple[str, ...]:
        """Tuple of report image locations."""
        return tuple(str(p) for p in self._report_images)

    def add_measurement(self, measurement: Measurement) -> None:
        """Add a measurement to the results."""
        self._measurements.append(measurement)

    def add_report_image(self, image_path: str | Sequence[str]) -> None:
        """Add a report image location to the report_images."""
        if isinstance(image_path, Sequence) and not isinstance(
            image_path,
            str,
        ):
            paths = image_path
        else:
            paths = [image_path]
        self._report_images += paths

    def get_measurement(
        self,
        name: str | None = None,
        measurement_type: str | None = None,
        subtype: str | None = None,
        description: str | None = None,
        unit: str | None = None,
    ) -> list[Measurement]:
        """Get the measurement(s) that match fields."""
        return [
            m
            for m in self.measurements
            if (
                (name is None or m.name == name)
                and (measurement_type is None or m.type == measurement_type)
                and (subtype is None or m.subtype == subtype)
                and (description is None or m.description in description)
                and (unit is None or m.unit == unit)
            )
        ]

    def filtered(self, level: str) -> Result:
        """Filter the measurements based on the specified level."""
        if level == "all":
            return self

        if level not in get_args(MEASUREMENT_VISIBILITY):
            raise InvalidMeasurementVisibilityError(level)

        new_result = Result(
            self.task,
            self.desc,
            self.files,
            _load_metadata=False,
        )

        new_result.metadata = self.metadata
        for img in self.report_images:
            new_result.add_report_image(img)

        for m in self.measurements:
            if m.visibility == level:
                new_result.add_measurement(m)
        return new_result

    def to_dict(self) -> dict[str, Any]:
        """Return dict."""
        # JsonSerializableMixin  doesn't include properties
        base = super().to_dict()

        # Add properties to dict.
        base["measurements"] = [m.to_dict() for m in self.measurements]
        base["report_images"] = list(self.report_images)
        base["metadata"] = self.metadata.to_dict()

        return base

    def __repr__(self) -> str:
        """Wrap around the to_dict method."""
        attrs = ", ".join(
            f"{key}={value!r}" for key, value in self.to_dict().items()
        )
        return f"{self.__class__.__name__}({attrs})"


#############################
### Task Specific Classes ###
#############################


########
# LCOD #
########


@dataclass
class LowContrastObject:
    """Class for the Low Contrast Present within each spoke."""

    x: int
    y: int
    diameter: float
    value: float | None = None
    detected: bool = False


@dataclass
class Spoke:
    """Spoke radiating from the center to the edge of the LCOD disk.

    Also contains the LowContrastObjects along the spoke.
    """

    cx: float
    cy: float
    theta: float  # Degrees

    diameter: float

    # Distance from the center (mm)
    dist: tuple[float, float, float] = (12.5, 25.0, 38.5)

    # Spoke length (mm)
    length: float = 43.0

    passed: bool = False

    objects: Sequence[LowContrastObject] = field(init=False)

    def __post_init__(self) -> None:
        """Initialise the objects within the spoke."""
        # y increases as the you down the image
        # hence the minus sign is so that theta = 0
        # corresponds to the object being visually
        # above the center rather than below.
        logger.debug("Creating objects along spoke angle: %f", self.theta)
        self.objects = tuple(
            LowContrastObject(
                x=self.cx + d * np.sin(np.deg2rad(self.theta)),
                y=self.cy - d * np.cos(np.deg2rad(self.theta)),
                diameter=self.diameter,
            )
            for d in self.dist
        )

    def __len__(self) -> int:
        """Objects within the spoke."""
        return len(self.objects)

    def __iter__(self) -> Iterator[LowContrastObject]:
        """Iterate over the low contrast objects."""
        return iter(self.objects)

    def profile(
        self,
        dcm: pydicom.FileDataset,
        size: int = 128,
        offset: tuple[float, float] = (0.0, 0.0),
        *,
        return_object_mask: bool = False,
        return_coords: bool = False,
    ) -> tuple[np.ndarray]:
        """Return the 1D profile of the DICOM for the spoke."""
        px_x, px_y = get_pixel_size(dcm)
        if px_x != px_y:
            msg = "Only square pixels are supported"
            logger.critical("%s but got (%f, %f)", msg, px_x, px_y)
            raise ValueError(msg)
        r_coords = np.linspace(
            0,
            self.length / px_x,
            size,
            endpoint=False,
        )
        theta_coords = np.zeros_like(r_coords) + self.theta

        x_coords = (self.cx + offset[1]) / px_x + r_coords * np.sin(
            np.deg2rad(theta_coords)
        )
        y_coords = (self.cy + offset[0]) / px_y - r_coords * np.cos(
            np.deg2rad(theta_coords)
        )

        profile = sp.ndimage.map_coordinates(
            dcm.pixel_array,
            [y_coords.ravel(), x_coords.ravel()],
            order=1,
        )

        rtn = [profile]

        if return_coords:
            rtn.append(np.array((x_coords, y_coords)))

        if return_object_mask:
            object_mask = np.zeros((size, len(self)), dtype=bool)
            r = np.linspace(0, self.length, size)
            for i, obj in enumerate(self):
                obj_r_pos = np.sqrt(
                    (obj.x - self.cx) ** 2 + (obj.y - self.cy) ** 2,
                )
                object_mask[:, i] = np.abs(r - obj_r_pos) <= obj.diameter / 4

            rtn.append(object_mask)

        if len(rtn) == 1:
            return rtn[0]

        return tuple(rtn)


@dataclass
class LCODTemplate:
    """Template for the LCOD object."""

    # Position of the center of the LCOD on the image.
    cx: float
    cy: float

    # Angle of rotation for the 1st spoke (degrees)
    theta: float

    # Diameters of each low contrast object (mm)
    # Starting with the largest spoke and moving clockwise.
    diameters: tuple[
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
    ] = (
        7.00,
        6.39,
        5.78,
        5.17,
        4.55,
        3.94,
        3.33,
        2.72,
        2.11,
        1.50,
    )

    @functools.cached_property
    def spokes(self) -> Sequence[Spoke]:
        """Position of each of the low contrast object spokes."""
        return self._calc_spokes(
            self.cx,
            self.cy,
            self.theta,
            self.diameters,
        )

    @staticmethod
    @functools.lru_cache(maxsize=10)
    def _calc_spokes(
        cx: float,
        cy: float,
        theta: float,
        diameters: tuple[float],
    ) -> tuple[Spoke, ...]:
        # Spokes start with the largest and decrease in size clockwise
        return tuple(
            Spoke(cx, cy, theta + i * (360 / len(diameters)), d)
            for i, d in enumerate(diameters)
        )

    def mask(
        self,
        dcm: pydicom.FileDataset,
        offset: tuple[float, float] = (0.0, 0.0),
        *,
        subset: str = "all",
        warn_if_object_out_of_bounds: bool = False,
    ) -> np.ndarray:
        """Mask the DICOM pixel array from the Spoke geometry.

        DICOM file is used for relating pixel geometry to
        the geometry of the acquisition.

        Similarly, offset (y, x) is used to account for cropped images.
        """
        mask = np.zeros_like(dcm.pixel_array, dtype=np.bool)

        # Convert from real coordinates to pixel coordinates
        dx, dy = get_pixel_size(dcm)
        y_grid, x_grid = np.meshgrid(
            *[
                np.linspace(v0, v0 + s * dv, num=s, endpoint=False)
                for v0, dv, s in zip(
                    offset,
                    (dy, dx),
                    mask.shape,
                )
            ],
            indexing="ij",
        )

        for sidx, spoke in enumerate(self.spokes):
            for oidx, obj in enumerate(spoke):
                is_object = (y_grid - obj.y) ** 2 + (x_grid - obj.x) ** 2 <= (
                    obj.diameter / 2
                ) ** 2
                # Compare actual to measured area to check if object is on the
                # grid.
                if warn_if_object_out_of_bounds:
                    mask_area = 2 * np.sum(is_object) * (dx * dy)
                    obj_area = np.pi * (obj.diameter / 2) ** 2
                    if not np.isclose(
                        mask_area,
                        obj_area,
                        rtol=1e-1,
                        atol=dx * dy,
                    ):
                        logger.warning(
                            "Object %d in spoke %d is out of bounds.\n"
                            "File: %s\n"
                            "Object area:\t%f\nMask area:\t%f\n"
                            "Object:\n\tCenter: (%f, %f)\n\tDiameter: %f\n"
                            "Image:\n\tPixel size: (%f, %f)\n\tShape: (%d, %d)",
                            oidx,
                            sidx,
                            dcm.filename,
                            obj_area,
                            mask_area,
                            obj.x,
                            obj.y,
                            obj.diameter,
                            dx,
                            dy,
                            *mask.shape,
                        )
                match subset:
                    case "all":
                        object_considered_for_mask = True
                    case "passed":
                        object_considered_for_mask = obj.detected
                    case "failed":
                        object_considered_for_mask = not obj.detected
                    case _:
                        logger.warning(
                            "Unrecognised mask subset %s."
                            " Object included in mask.",
                            subset,
                        )
                        object_considered_for_mask = True

                if object_considered_for_mask:
                    mask |= is_object
        return mask


@dataclass
class StatsParameters:
    """Dataclass for the stats parameters and associated p-values."""

    p_vals: list[tuple] = field(default_factory=list)
    params: list[tuple] = field(default_factory=list)

    @property
    def p_vals_all(self) -> list:
        """All the p-values as a flat list."""
        p_vals_all = []
        for p in self.p_vals:
            p_vals_all += list(p)
        return p_vals_all

    @property
    def params_all(self) -> list:
        """All the parameters as a flat list."""
        params_all = []
        for p in self.params:
            params_all += list(p)
        return params_all


@dataclass(frozen=True)
class FailedStatsModel:
    """Dataclass for the failed stats model."""

    @property
    def pvalues(self) -> np.ndarray:
        """Return the p-values - all ones."""
        return np.ones(3)

    @property
    def params(self) -> np.ndarray:
        """Return the p-values - all ones."""
        return np.ones(3)


@dataclass
class SpokeReportData:
    """Container for spoke analysis data used in reporting."""

    spoke_id: int
    profile: np.ndarray
    detrended: np.ndarray
    trend: np.ndarray
    object_mask: np.ndarray
    x_coords: np.ndarray | None
    y_coords: np.ndarray | None
    p_vals: np.ndarray
    params: np.ndarray
    detected: list[bool]
    objects: list[Any]  # Object info from template



"""Output formatters from dict-like objects."""

# ruff: noqa: ANN401

from __future__ import annotations

# Type Checking
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping

    from hazenlib.types import Result

# Python imports
import csv
import logging
import sys
from pathlib import Path
from typing import Any, Literal, TextIO, TypeAlias, get_args

# Local imports
from hazenlib.constants import MEASUREMENT_VISIBILITY
from hazenlib.exceptions import InvalidMeasurementVisibilityError

logger = logging.getLogger(__name__)

# ----------------------------------------------------------------------
# Public type alias for the format argument
# ----------------------------------------------------------------------

Format: TypeAlias = Literal["json", "csv", "tsv"]

# ----------------------------------------------------------------------
# Public API
# ----------------------------------------------------------------------


def write_result(
    data: Result,
    fmt: Format,
    path: str | Path = "-",
    level: str = "all",
) -> None:
    r"""Serialise data to a format and write the output.

    Serialize *data* into the requested *fmt* and write it either to
    ``stdout`` (default) or to a file.

    Parameters
    ----------
    data:
        Mapping that will be serialised.
    fmt:
        Output format. ``json`` is the default; other choices are
        ``csv``, ``tsv``.
    path:
        Path to the destination file.  ``"-"`` (the default) means
        write to ``sys.stdout``.  Any other value is interpreted as a
        filesystem path; the file is opened in text mode with
        ``newline=\"\"`` for CSV/TSV compatibility.
    level:
        Filter the measurements by a specific level.

    Raises
    ------
    ValueError
        If *fmt* is not one of the supported literals.

    """
    try:
        data = data.filtered(level)
    except InvalidMeasurementVisibilityError:
        logger.warning(
            "Unknown measurement filter level %s :"
            " allowed options are %s - defaulting to 'all'",
            level,
            [*get_args(MEASUREMENT_VISIBILITY), "all"],
        )

    if path == "-":
        _format_results(data, fmt, sys.stdout)
        return
    with Path(path).open("a", newline="") as fp:
        write_header = not Path(path).exists() or not Path(path).stat().st_size
        _format_results(data, fmt, fp, write_header=write_header)


# ----------------------------------------------------------------------
# Private API
# ----------------------------------------------------------------------


def _to_python_scalar(value: Any) -> Any:
    try:
        return value.item()
    except AttributeError:
        return value


def _format_results(
    data: Mapping[str, Any],
    fmt: Format,
    out_fh: TextIO,
    *,
    write_header: bool = True,
) -> None:
    """Dispatcher that writes *data* to *out_fh* in the requested *fmt*.

    The function assumes *out_fh* is already opened and will **not** close it.
    """
    match fmt:
        case "json":
            print(data.to_json(), file=out_fh)
        case "csv":
            _write_csv(data, out_fh, delimiter=",", write_header=write_header)
        case "tsv":
            _write_csv(data, out_fh, delimiter="\t", write_header=write_header)
        case _:
            msg = f"Unrecognised format {fmt!r}"
            logger.critical(msg)
            raise ValueError(msg)


def _write_csv(
    data: Mapping[str, Any],
    out_fh: TextIO,
    *,
    delimiter: str,
    write_header: bool = True,
) -> None:
    """Write a flat mapping as a two-row CSV/TSV file."""
    rows = _build_rows(data)
    if not rows:
        logger.warning("No rows generated from data.")
        logger.debug(str(data))
        return

    writer = csv.DictWriter(
        out_fh,
        fieldnames=rows[0].keys(),
        delimiter=delimiter,
    )
    if write_header:
        writer.writeheader()
    for row in rows:
        writer.writerow(row)


def _build_rows(data: Mapping[str, Any]) -> list[dict]:
    """Convert nested data to flat rows.

    Convert the nested result dictionary
    into a list of flat rows suitable for CSV.
    """
    d = data.to_dict()
    measurements = d.pop("measurements", {})
    metadata = d.pop("metadata", {})
    return [{**m, **d, **metadata} for m in measurements]



# Python imports
import logging
import os.path
import unittest

# Module imports
import matplotlib as mpl
import numpy as np

# Local imports
from hazenlib.tasks.snr_map import SNRMap
from tests import TEST_DATA_DIR

logger = logging.getLogger(__name__)


class TestSnrMap(unittest.TestCase):
    siemens_1 = [
        os.path.join(TEST_DATA_DIR, "snr", "Siemens", "tra_250_2meas_1.IMA")
    ]

    ROI_CORNERS_TEST = [
        np.array([114, 121]),
        np.array([74, 81]),
        np.array([154, 81]),
        np.array([74, 161]),
        np.array([154, 161]),
    ]
    IMAGE_CENTRE_TEST = np.array([123.7456188, 131.21848254])

    def __init__(self, methodName: str = ...):
        super().__init__(methodName)

    def setUp(self):
        self.snr_map_task = SNRMap(input_data=self.siemens_1, report=True)
        self.results = self.snr_map_task.run()
        self.original, self.smoothed, self.noise = self.snr_map_task.smooth(
            dcm=self.snr_map_task.single_dcm,
            kernel=self.snr_map_task.kernel_size,
        )
        self.image_centre, self.roi_corners = self.snr_map_task.get_rois(
            self.smoothed
        )
        self.snr = self.snr_map_task.calc_snr(
            self.original, self.noise, self.roi_corners
        )
        self.snr_map = self.snr_map_task.calc_snr_map(
            self.original, self.noise
        )
        self.detailed_fig = self.snr_map_task.plot_detailed(
            self.original,
            self.smoothed,
            self.noise,
            self.snr,
            self.snr_map,
            self.image_centre,
            self.roi_corners,
        )
        self.summary_fig = self.snr_map_task.plot_summary(
            self.snr_map, self.original, self.roi_corners
        )

    def test_snr_value(self):
        np.testing.assert_allclose(
            192.88188017908504,
            self.results.get_measurement(
                name="SNR",
                measurement_type="measured",
                subtype="smoothing",
            )[0].value,
            2,
        )

    def test_smooth(self):
        np.testing.assert_allclose(self.original.cumsum().sum(), 1484467722691)
        np.testing.assert_allclose(
            self.smoothed.cumsum().sum(), 1484468146211.5635
        )
        np.testing.assert_allclose(abs(self.noise).sum(), 2147755.9753086423)

    def test_get_rois(self):
        np.testing.assert_allclose(self.roi_corners, self.ROI_CORNERS_TEST)
        np.testing.assert_allclose(self.image_centre, self.IMAGE_CENTRE_TEST)
        assert self.snr_map_task.mask.sum() == 29444

    def test_calc_snr(self):
        np.testing.assert_approx_equal(self.snr, 192.8818801790859)

    def test_calc_snr_map(self):
        snr_map_cumsum = self.snr_map.cumsum().sum()

        fixed_val = 128077116718.40483
        logger.debug(
            "\ntest_calc_snr_map.py::TestCalcSnrMap::test_calc_snr_map"
            "new_release_value: %s\n"
            "fixed_value: %s,",
            snr_map_cumsum,
            fixed_val,
        )

        np.testing.assert_allclose(snr_map_cumsum, fixed_val)

    def test_plot_detailed(self):
        # Just check a valid figure handle is returned
        assert isinstance(self.detailed_fig, mpl.figure.Figure)

    def test_plot_summary(self):
        # Just check a valid figure handle is returned
        assert isinstance(self.summary_fig, mpl.figure.Figure)


if __name__ == "__main__":
    unittest.main()



"""Test suite for the Geometric Accuracy."""

# ruff: noqa: T201 S101

import logging
import pathlib
import unittest

import numpy as np
from hazenlib.tasks.acr_geometric_accuracy import ACRGeometricAccuracy
from hazenlib.utils import get_dicom_files

from tests import TEST_DATA_DIR, TEST_REPORT_DIR

logger = logging.getLogger(__name__)


###########
# Siemens #
###########


class TestACRGeometricAccuracySiemens(unittest.TestCase):
    """Geometric accuracy for legacy Siemens test data."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens")
    L1 = (191.41, 187.5)
    L5 = (191.41, 187.5, 191.41, 191.41)
    distortion_metrics = (0.11, 2.5, 0.97)

    def setUp(self) -> None:
        """Set up the test object."""
        input_files = get_dicom_files(self.ACR_DATA)

        self.acr_geometric_accuracy_task = ACRGeometricAccuracy(
            input_data=input_files,
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )

        self.dcm_1 = self.acr_geometric_accuracy_task.ACR_obj.slice_stack[0]
        self.dcm_5 = self.acr_geometric_accuracy_task.ACR_obj.slice_stack[4]

    def test_geometric_accuracy_slice_1(self) -> None:
        """Test Geometric Accuracy with Slice 1."""
        slice1_vals = self.acr_geometric_accuracy_task.get_geometric_accuracy(
            0
        )
        slice1_vals = np.round(slice1_vals, 2)

        logger.info(
            "Slice 1:\nnew_release: %s\nfixed value: %s",
            slice1_vals,
            self.L1,
        )

        assert (slice1_vals == self.L1).all()

    def test_geometric_accuracy_slice_5(self) -> None:
        """Test Geometric Accuracy with Slice 5."""
        slice5_vals = np.array(
            self.acr_geometric_accuracy_task.get_geometric_accuracy(4),
        )

        slice5_vals = np.round(slice5_vals, 2)

        logger.info(
            "Slice 5:\nnew_release: %s\nfixed value: %s",
            slice5_vals,
            self.L5,
        )

        assert (slice5_vals == self.L5).all()

    def test_distortion_metrics(self) -> None:
        """Test distortion metrics."""
        metrics = np.array(
            self.acr_geometric_accuracy_task.get_distortion_metrics(
                self.L1 + self.L5,
            ),
        )
        metrics = np.round(metrics, 2)
        assert (metrics == self.distortion_metrics).all()


class TestACRGeometricAccuracySiemensAeraT1(TestACRGeometricAccuracySiemens):
    """Test Data for the Siemens_Aera_1.5T_T1 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens_Aera_1.5T_T1")
    L1 = (189.45, 190.43)
    L5 = (190.43, 191.41, 189.45, 189.45)
    distortion_metrics = (0.1, 1.41, 0.38)


class TestACRGeometricAccuracySiemensAeraT2(
    TestACRGeometricAccuracySiemensAeraT1,
):
    """Test Data for the Siemens_Aera_1.5T_T2 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens_Aera_1.5T_T2")


class TestACRGeometricAccuracySiemensMagnetomSkyraT1(
    TestACRGeometricAccuracySiemens,
):
    """Test Data for the Siemens_MagnetomSkyra_3T_T1 dataset."""

    ACR_DATA = (
        pathlib.Path(TEST_DATA_DIR) / "acr" / "Siemens_MagnetomSkyra_3T_T1"
    )
    L1 = (192.38, 192.38)
    L5 = (191.41, 191.41, 191.41, 191.41)
    distortion_metrics = (1.73, 2.38, 0.24)


class TestACRGeometricAccuracySiemensMagnetomSkyraT2(
    TestACRGeometricAccuracySiemensMagnetomSkyraT1,
):
    """Test Data for the Siemens_MagnetomSkyra_3T_T2 dataset."""

    ACR_DATA = (
        pathlib.Path(TEST_DATA_DIR) / "acr" / "Siemens_MagnetomSkyra_3T_T2"
    )


class TestACRGeometricAccuracySiemensSolaT1(TestACRGeometricAccuracySiemens):
    """Test Data for the Siemens_Sola_1.5T_T1 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens_Sola_1.5T_T1")
    L1 = (190.43, 189.45)
    L5 = (189.45, 190.43, 191.41, 189.45)
    distortion_metrics = (0.1, 1.41, 0.38)


class TestACRGeometricAccuracySiemensSolaT2(
    TestACRGeometricAccuracySiemensSolaT1,
):
    """Test Data for the Siemens_Sola_1.5T_T2 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens_Sola_1.5T_T2")
    L5 = (189.45, 190.43, 189.45, 189.45)
    distortion_metrics = (-0.22, 0.55, 0.24)


class TestACRGeometricAccuracySiemensSolaFit(TestACRGeometricAccuracySiemens):
    """Test Data for the Siemens Sola Fit dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "SiemensSolaFit")
    L1 = (190.43, 190.43)
    L5 = (189.45, 190.43, 191.41, 191.41)
    distortion_metrics = (0.59, 1.41, 0.35)


######
# GE #
######


class TestACRGeometricAccuracyGE(TestACRGeometricAccuracySiemens):
    """Test Data for the GE dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE")
    L1 = (190.93, 188.9)
    L5 = (190.42, 189.41, 190.43, 189.41)
    distortion_metrics = (-0.08, 1.1, 0.38)


class TestACRGeometricAccuracyGEArtistT1(TestACRGeometricAccuracyGE):
    """Test Data for the GE Artist 1.5T T1 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE_Artist_1.5T_T1")
    L1 = (190.44, 190.44)
    L5 = (190.44, 190.44, 189.46, 191.41)
    distortion_metrics = (0.44, 1.41, 0.30)


class TestACRGeometricAccuracyGEArtistT2(TestACRGeometricAccuracyGEArtistT1):
    """Test Data for the GE Artist 1.5T T2 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE_Artist_1.5T_T2")
    L5 = (190.44, 190.44, 189.46, 190.44)
    distortion_metrics = (0.28, 0.54, 0.19)


class TestACRGeometricAccuracyGEMR450WT1(TestACRGeometricAccuracyGE):
    """Test Data for the GE MR450W 1.5T T1 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE_MR450W_1.5T_T1")
    L1 = (188.48, 188.48)
    L5 = (188.48, 188.48, 190.44, 191.41)
    distortion_metrics = (-0.71, 1.52, 0.63)


class TestACRGeometricAccuracyGEMR450WT2(TestACRGeometricAccuracyGEMR450WT1):
    """Test Data for the GE MR450W 1.5T T2 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE_MR450W_1.5T_T2")
    L1 = (189.46, 188.48)
    L5 = (188.48, 188.48, 191.41, 189.46)
    distortion_metrics = (-0.71, 1.52, 0.55)


class TestACRGeometricAccuracyGESignaT1(TestACRGeometricAccuracyGE):
    """Test Data for the GE Signa 3T T1 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE_Signa_3T_T1")
    L1 = (189.46, 189.46)
    L5 = (190.44, 188.48, 191.41, 191.41)
    distortion_metrics = (0.11, 1.52, 0.57)


class TestACRGeometricAccuracyGESignaT2(TestACRGeometricAccuracyGESignaT1):
    """Test Data for the GE Signa 3T T2 dataset."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE_Signa_3T_T2")
    L1 = (189.46, 189.46)
    L5 = (190.44, 188.48, 191.41, 191.41)
    distortion_metrics = (0.11, 1.52, 0.57)


###########
# Philips #
###########


class TestACRGeometricAccuracyPhilipsAchieva(TestACRGeometricAccuracySiemens):
    """Test data for the Philips Achieva."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "PhilipsAchieva")
    L1 = 190.43, 189.45
    L5 = 190.43, 189.45, 189.45, 189.45
    distortion_metrics = (-0.22, 0.55, 0.24)



import os
import pathlib
import unittest
import pytest

from hazenlib.tasks.uniformity import Uniformity
from tests import TEST_DATA_DIR, TEST_REPORT_DIR


class TestUniformity(unittest.TestCase):
    UNIFORMITY_DATA = pathlib.Path(TEST_DATA_DIR / "uniformity")
    IPEM_HORIZONTAL = 1.0
    IPEM_VERTICAL = 0.98125

    def setUp(self):
        self.uniformity_task = Uniformity(
            input_data=[os.path.join(self.UNIFORMITY_DATA, "axial_oil.IMA")],
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )

    def test_uniformity(self):
        results = self.uniformity_task.run()
        horizontal_ipem = results.get_measurement(
            name="Uniformity",
            measurement_type="measured",
            subtype="Horizontal",
        )[0].value
        vertical_ipem = results.get_measurement(
            name="Uniformity",
            measurement_type="measured",
            subtype="Vertical",
        )[0].value

        print("\ntest_uniformity.py::TestUniformity::test_uniformity")
        print("new_release_value:", vertical_ipem)
        print("fixed_value:", self.IPEM_VERTICAL)

        assert horizontal_ipem == pytest.approx(
            self.IPEM_HORIZONTAL, abs=0.005
        )
        assert vertical_ipem == pytest.approx(self.IPEM_VERTICAL, abs=0.005)


class TestSagUniformity(TestUniformity):
    IPEM_HORIZONTAL = 0.46875
    IPEM_VERTICAL = 0.5125

    def setUp(self):
        self.uniformity_task = Uniformity(
            input_data=[os.path.join(self.UNIFORMITY_DATA, "sag.dcm")],
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )


class TestCorUniformity(TestUniformity):
    IPEM_HORIZONTAL = 0.35
    IPEM_VERTICAL = 0.45

    def setUp(self):
        self.uniformity_task = Uniformity(
            input_data=[os.path.join(self.UNIFORMITY_DATA, "cor.dcm")],
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )



import unittest
import pathlib

import numpy as np

# import hazenlib.slice_width as hazen_slice_width
from tests import TEST_DATA_DIR, TEST_REPORT_DIR
from hazenlib.tasks.slice_width import SliceWidth
from hazenlib.utils import Rod
import os


class TestSliceWidth(unittest.TestCase):
    SLICE_WIDTH_DATA = pathlib.Path(TEST_DATA_DIR / "slicewidth")

    """
    Notes
    -----
    The rod indices are ordered as:
        789
        456
        123 
    """
    matlab_rods = [
        Rod(71.2857, 191.5000),
        Rod(130.5000, 190.5000),
        Rod(190.6000, 189.4000),
        Rod(69.5000, 131.5000),
        Rod(128.7778, 130.5000),
        Rod(189.1111, 129.2778),
        Rod(69.0000, 71.5000),
        Rod(128.1176, 70.4118),
        Rod(188.5000, 69.2222),
    ]

    rods = [
        Rod(70.26906602941604, 190.52291430040833),
        Rod(129.38344648450575, 189.5252799358382),
        Rod(189.6494724536544, 188.32774808447635),
        Rod(68.53084886954112, 130.56732921648214),
        Rod(127.86240947286896, 129.4605302262616),
        Rod(188.01124565987345, 128.2832650316875),
        Rod(67.97926729691507, 70.61103769200058),
        Rod(127.2060664869085, 69.42672715143607),
        Rod(187.49797283835656, 68.2890101413575),
    ]

    DISTANCES = ([119.333, 119.632, 119.522], [120.022, 120.112, 120.196])
    DIST_CORR_COEFF = {"top": 0.9965, "bottom": 0.9957}
    ROD_DIST = (0.13, 0.07)

    TOP_CENTRE = 102
    BOTTOM_CENTRE = 162
    MATLAB_PROFILE = [
        693,
        694.75,
        688.6,
        685.05,
        682.2,
        677.65,
        677.35,
        675.5,
        668,
        665.9,
        663.05,
        662.05,
        662.1,
        657.75,
        653.65,
        654.75,
        650.1,
        648.65,
        646.1,
        645.9,
        643.1,
        639.4,
        641.35,
        639.1,
        635.95,
        633.75,
        633.9,
        630.7,
        627.25,
        628.2,
        628.05,
        623.4,
        624.35,
        622.4,
        621.15,
        622.45,
        615,
        615.8,
        611.15,
        605.2,
        602.15,
        595.75,
        582.75,
        573.5,
        561.25,
        550.4,
        538.25,
        524.9,
        511.65,
        499.9,
        489.7,
        487.3,
        477,
        476.1,
        469.55,
        466.9,
        467.7,
        466,
        465.05,
        468,
        464.7,
        467.9,
        464.65,
        468.85,
        467.55,
        468,
        469.05,
        473.25,
        479.6,
        483.8,
        492.3,
        500,
        512.5,
        523,
        535.95,
        548.3,
        561.8,
        571.95,
        585.35,
        593.5,
        601.55,
        605.45,
        609.3,
        612.75,
        617.1,
        618.8,
        620.35,
        623.15,
        621.7,
        623.35,
        625.25,
        628.4,
        628.8,
        631.55,
        632.55,
        633.4,
        634.35,
        638.1,
        639.35,
        640.3,
        641.75,
        646.7,
        648.1,
        647.2,
        648.95,
        650.2,
        653.95,
        659.55,
        660.8,
        661.95,
        663,
        665.6,
        666.15,
        670.75,
        672.6,
        674.15,
        675.9,
        677.5,
        682.45,
        684.3,
    ]

    BLINE_TOP = 0.0215
    BLINE_BOT = 0.0239

    TRAP_FIT_COEFF_TOP = [47, 55, 153, 178]
    TRAP_FIT_COEFF_BOT = [47, 55, 164, 167]

    MATLAB_TRAP_FIT_COEFF = [50, 54, 153, 170, -111.7920]
    MATLAB_BLINE_FIT_COEFF = [0.0216, -2.9658, 602.2568]
    MATLAB_TRAP_FIT_COEFF_BOT = [
        55.0000,
        60.0000,
        155.0000,
        152.0000,
        -136.6194,
    ]
    MATLAB_BLINE_FIT_COEFF_BOT = [0.0239, -2.9389, 694.9520]

    SW_MATLAB = 5.48

    def setUp(self):
        # self.file = str(self.SLICE_WIDTH_DATA / 'SLICEWIDTH' / 'ANNUALQA.MR.HEAD_GENERAL.tra.slice_width.IMA')
        # self.dcm = dcmread(self.file)
        self.slice_width = SliceWidth(
            input_data=[
                os.path.join(
                    self.SLICE_WIDTH_DATA,
                    "SLICEWIDTH",
                    "ANNUALQA.MR.HEAD_GENERAL.tra.slice_width.IMA",
                )
            ],
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )

    def test_get_rods(self):
        arr = self.slice_width.single_dcm.pixel_array
        rods, _ = self.slice_width.get_rods(arr)
        # print("rods")
        # print(rods)
        for n in range(len(rods)):
            np.testing.assert_almost_equal(
                self.rods[n].centroid, rods[n].centroid, 3
            )

    def test_get_rod_distances(self):
        # From MATLAB Rods
        distances = self.slice_width.get_rod_distances(self.matlab_rods)
        np.testing.assert_almost_equal(distances, self.DISTANCES, 3)

    def test_get_rod_distortion_correction_coefficients(self):
        distances = self.slice_width.get_rod_distances(self.matlab_rods)
        # print("rod distortion correction coefficient")
        # print(hazen_slice_width.get_rod_distortion_correction_coefficients(distances[0], self.dcm.PixelSpacing[0]))
        dist_corr_coeff = (
            self.slice_width.get_rod_distortion_correction_coefficients(
                distances[0]
            )
        )

        np.testing.assert_almost_equal(
            dist_corr_coeff["top"], self.DIST_CORR_COEFF["top"], 3
        )
        np.testing.assert_almost_equal(
            dist_corr_coeff["bottom"], self.DIST_CORR_COEFF["bottom"], 3
        )

    def test_rod_distortions(self):
        horz_dist, vert_dist = self.slice_width.get_rod_distances(
            self.matlab_rods
        )
        (
            horizontal_distortion,
            vertical_distortion,
        ) = self.slice_width.get_rod_distortions(horz_dist, vert_dist)
        # print("rod distortion")
        # print(horizontal_distortion, vertical_distortion)
        assert (
            round(horizontal_distortion, 2),
            round(vertical_distortion, 2),
        ) == self.ROD_DIST

    def test_get_ramp_profiles(self):
        ramp_profiles = self.slice_width.get_ramp_profiles(
            self.slice_width.single_dcm.pixel_array, self.matlab_rods
        )
        bottom_profiles = ramp_profiles["bottom"]
        mean_bottom_profile = np.mean(bottom_profiles, axis=0).tolist()
        # print("bottom centre ramp profile")
        # print(ramp_profiles["bottom-centre"])
        # print("top centre ramp profile")
        # print(ramp_profiles["top-centre"])
        # print("mean bottom profile")
        # print(mean_bottom_profile)
        assert ramp_profiles["bottom-centre"] == self.BOTTOM_CENTRE
        assert ramp_profiles["top-centre"] == self.TOP_CENTRE
        assert mean_bottom_profile == self.MATLAB_PROFILE

    def test_baseline_correction(self):
        # matlab top 0.0215   -2.9668  602.4568
        # matlab bottom [0.0239, -2.9349,  694.9520]

        ramps = self.slice_width.get_ramp_profiles(
            self.slice_width.single_dcm.pixel_array, self.matlab_rods
        )

        top_mean_ramp = np.mean(ramps["top"], axis=0)
        top_coefficients = list(
            self.slice_width.baseline_correction(
                top_mean_ramp, sample_spacing=0.25
            )["f"]
        )
        # print("top bline  corr coeff")
        # print(round(top_coefficients[0], 4))
        assert round(top_coefficients[0], 4) == self.BLINE_TOP

        bottom_mean_ramp = np.mean(ramps["bottom"], axis=0)
        bottom_coefficients = list(
            self.slice_width.baseline_correction(
                bottom_mean_ramp, sample_spacing=0.25
            )["f"]
        )
        # print("bottom bline corr coeff")
        # print(round(bottom_coefficients[0], 4))
        assert round(bottom_coefficients[0], 4) == self.BLINE_BOT

    def test_trapezoid(self):
        """
        Notes
        -----
        variables from one iteration of the original matlab script
        n_ramp, n_plateau, n_left_baseline, n_right_baseline, plateau_amplitude =
        55.0000   58.0000  156.0000  153.0000 -136.6194 and fwhm 113
        """

        assert (
            self.slice_width.trapezoid(55, 58, 156, 153, -136.6194)[1] == 113
        )

    def test_get_initial_trapezoid_fit_and_coefficients(self):
        """
        Notes
        -----
        Trapezoid_Fit_Coefficients_Initial = 48.0000   56.0000  153.0000  172.0000 -116.4920
        fwhm = 104

        """
        sample_spacing = 0.25
        slice_thickness = self.slice_width.single_dcm.SliceThickness
        ramps = self.slice_width.get_ramp_profiles(
            self.slice_width.single_dcm.pixel_array, self.matlab_rods
        )
        top_mean_ramp = np.mean(ramps["top"], axis=0)
        bottom_mean_ramp = np.mean(ramps["bottom"], axis=0)
        ramps_baseline_corrected = {
            "top": self.slice_width.baseline_correction(
                top_mean_ramp, sample_spacing
            ),
            "bottom": self.slice_width.baseline_correction(
                bottom_mean_ramp, sample_spacing
            ),
        }

        (
            trapezoid_fit,
            trapezoid_fit_coefficients,
        ) = self.slice_width.get_initial_trapezoid_fit_and_coefficients(
            ramps_baseline_corrected["top"]["profile_corrected_interpolated"],
            slice_thickness,
        )
        # print("trap fit coeff top initial")
        # print(trapezoid_fit_coefficients[:4])
        assert trapezoid_fit_coefficients[:4] == self.TRAP_FIT_COEFF_TOP

        (
            trapezoid_fit,
            trapezoid_fit_coefficients,
        ) = self.slice_width.get_initial_trapezoid_fit_and_coefficients(
            ramps_baseline_corrected["bottom"][
                "profile_corrected_interpolated"
            ],
            slice_thickness,
        )
        # print("trap fit coeff bottom initial")
        # print(trapezoid_fit_coefficients[:4])
        assert trapezoid_fit_coefficients[:4] == self.TRAP_FIT_COEFF_BOT

    def test_fit_trapezoid(self):
        sample_spacing = 0.25
        slice_thickness = self.slice_width.single_dcm.SliceThickness

        ramps = self.slice_width.get_ramp_profiles(
            self.slice_width.single_dcm.pixel_array, self.matlab_rods
        )

        top_mean_ramp = np.mean(ramps["top"], axis=0)
        bottom_mean_ramp = np.mean(ramps["bottom"], axis=0)
        ramps_baseline_corrected = {
            "top": self.slice_width.baseline_correction(
                top_mean_ramp, sample_spacing
            ),
            "bottom": self.slice_width.baseline_correction(
                bottom_mean_ramp, sample_spacing
            ),
        }
        (
            trapezoid_fit_coefficients,
            baseline_fit_coefficients,
        ) = self.slice_width.fit_trapezoid(
            profiles=ramps_baseline_corrected["top"],
            slice_thickness=slice_thickness,
        )
        # print("top trap fit coeff")
        # print(trapezoid_fit_coefficients)

        # print("top bline fit coeff")
        # print(baseline_fit_coefficients)

        # check top profile first
        matlab_trapezoid_fit_coefficients = self.MATLAB_TRAP_FIT_COEFF
        matlab_baseline_fit_coefficients = self.MATLAB_BLINE_FIT_COEFF

        for idx, value in enumerate(trapezoid_fit_coefficients):
            assert abs(value - matlab_trapezoid_fit_coefficients[idx]) <= 5

        for idx, value in enumerate(matlab_baseline_fit_coefficients):
            assert abs(value - matlab_baseline_fit_coefficients[idx]) <= 5
        # residual error is 3735.6

        ## check bottom profile now
        (
            trapezoid_fit_coefficients,
            baseline_fit_coefficients,
        ) = self.slice_width.fit_trapezoid(
            profiles=ramps_baseline_corrected["bottom"],
            slice_thickness=slice_thickness,
        )

        # print("bottom trap fit coeff")
        # print(trapezoid_fit_coefficients)
        # print("bottom bline fit coeff")
        # print(baseline_fit_coefficients)

        matlab_trapezoid_fit_coefficients = self.MATLAB_TRAP_FIT_COEFF_BOT
        matlab_baseline_fit_coefficients = self.MATLAB_BLINE_FIT_COEFF_BOT
        for idx, value in enumerate(trapezoid_fit_coefficients):
            assert abs(value - matlab_trapezoid_fit_coefficients[idx]) <= 5

        for idx, value in enumerate(matlab_baseline_fit_coefficients):
            assert abs(value - matlab_baseline_fit_coefficients[idx]) <= 5

    def test_slice_width(self):
        result = self.slice_width.run()
        slice_width_mm = result.get_measurement(name="SliceWidth")[0].value

        print("\ntest_slice_width.py::TestSliceWidth::test_slice_width")
        print("new_release_value:", slice_width_mm)
        print("fixed_value:", self.SW_MATLAB)

        assert abs(slice_width_mm - self.SW_MATLAB) < 0.1


class Test512Matrix(TestSliceWidth):
    SLICE_WIDTH_DATA = pathlib.Path(TEST_DATA_DIR / "slicewidth")

    """
    Notes
    -----
    The rod indices are ordered as:
        789
        456
        123
    """
    matlab_rods = [
        Rod(134.19422395407386, 376.3297092734241),
        Rod(255.5191606119133, 374.7890799840584),
        Rod(376.45583336255567, 373.7479398253673),
        Rod(133.73482492069846, 255.9592867481753),
        Rod(254.49068286441252, 254.33374023796569),
        Rod(375.15134878569734, 254.0979036468634),
        Rod(133.07301997760993, 136.86070859809843),
        Rod(253.53888071406627, 135.75358569280274),
        Rod(374.00513210724, 135.12370483098044),
    ]

    rods = [
        Rod(134.50715482125025, 376.39901305569873),
        Rod(255.46730860444976, 374.7645374009015),
        Rod(376.3508021809186, 373.6769436835617),
        Rod(133.58289022243835, 256.06611582258665),
        Rod(254.51562676421784, 254.61914241989848),
        Rod(375.3779965519839, 253.83655980142507),
        Rod(132.7380060479853, 137.19733440032368),
        Rod(253.45972063728271, 135.83081591213042),
        Rod(374.31705777525144, 135.3055747931858),
    ]

    DISTANCES = ([242.275, 241.424, 240.938], [239.472, 239.044, 238.637])
    DIST_CORR_COEFF = {"top": 1.0049, "bottom": 1.0077}
    ROD_DIST = (0.28, 0.17)

    TOP_CENTRE = 196
    BOTTOM_CENTRE = 316
    MATLAB_PROFILE = [
        7970.0,
        7912.7,
        7884.8,
        7858.8,
        7812.95,
        7759.8,
        7717.925,
        7686.575,
        7658.275,
        7625.375,
        7578.45,
        7522.275,
        7472.325,
        7439.35,
        7408.325,
        7364.875,
        7327.475,
        7311.0,
        7299.025,
        7265.475,
        7213.525,
        7166.7,
        7122.625,
        7070.875,
        7022.25,
        6998.95,
        6990.05,
        6965.325,
        6923.2,
        6891.6,
        6880.05,
        6856.425,
        6809.975,
        6770.95,
        6759.55,
        6752.15,
        6721.55,
        6682.175,
        6658.075,
        6642.175,
        6611.75,
        6562.325,
        6512.85,
        6485.575,
        6479.45,
        6478.3,
        6467.8,
        6439.45,
        6396.675,
        6351.825,
        6320.7,
        6309.525,
        6310.175,
        6303.0,
        6267.3,
        6214.1,
        6173.725,
        6167.0,
        6176.825,
        6171.4,
        6152.0,
        6146.4,
        6147.425,
        6127.85,
        6090.75,
        6065.175,
        6055.675,
        6031.15,
        5991.975,
        5977.125,
        5996.25,
        5998.8,
        5951.0,
        5895.325,
        5883.475,
        5903.65,
        5899.7,
        5861.825,
        5843.075,
        5860.825,
        5872.95,
        5857.625,
        5835.2,
        5828.275,
        5816.95,
        5783.475,
        5752.975,
        5753.025,
        5766.5,
        5768.75,
        5773.775,
        5794.775,
        5802.7,
        5757.25,
        5684.375,
        5638.1,
        5624.1,
        5593.575,
        5519.275,
        5428.525,
        5351.5,
        5274.45,
        5169.925,
        5045.75,
        4932.65,
        4839.15,
        4757.825,
        4684.975,
        4621.675,
        4577.3,
        4547.775,
        4529.55,
        4515.15,
        4495.7,
        4479.775,
        4484.425,
        4501.25,
        4487.35,
        4437.025,
        4392.075,
        4386.55,
        4401.175,
        4401.925,
        4386.75,
        4388.975,
        4402.075,
        4395.4,
        4376.25,
        4384.05,
        4419.075,
        4443.875,
        4436.9,
        4423.475,
        4437.25,
        4467.7,
        4485.775,
        4490.65,
        4499.55,
        4522.525,
        4556.525,
        4594.4,
        4634.55,
        4662.325,
        4658.45,
        4645.775,
        4675.325,
        4746.475,
        4802.05,
        4809.95,
        4813.95,
        4874.025,
        4968.925,
        5048.425,
        5119.975,
        5236.2,
        5398.2,
        5542.675,
        5632.375,
        5706.4,
        5813.375,
        5943.575,
        6056.1,
        6135.025,
        6193.7,
        6246.75,
        6291.075,
        6325.35,
        6353.55,
        6371.325,
        6382.85,
        6403.925,
        6445.9,
        6493.45,
        6522.275,
        6531.15,
        6535.175,
        6541.625,
        6544.275,
        6557.0,
        6593.725,
        6651.325,
        6703.0,
        6725.95,
        6732.325,
        6745.5,
        6774.625,
        6811.225,
        6843.275,
        6863.175,
        6874.65,
        6893.3,
        6934.7,
        6985.0,
        7019.45,
        7049.125,
        7095.475,
        7143.35,
        7165.65,
        7169.85,
        7197.3,
        7255.65,
        7304.35,
        7319.925,
        7333.55,
        7385.35,
        7466.05,
        7530.7,
        7559.15,
        7579.95,
        7614.85,
        7657.725,
        7699.05,
        7747.975,
        7805.1,
        7855.1,
        7888.925,
        7913.6,
        7947.425,
        7996.65,
        8051.65,
        8105.075,
        8160.575,
        8215.6,
        8258.975,
        8284.975,
        8312.85,
        8364.925,
        8438.1,
        8501.675,
        8547.175,
        8597.825,
        8665.9,
        8731.175,
        8776.475,
        8814.125,
        8864.925,
        8922.575,
        8972.85,
        9023.675,
        9093.4,
        9167.25,
        9216.7,
    ]

    BLINE_TOP = 0.1686
    BLINE_BOT = 0.2072

    TRAP_FIT_COEFF_TOP = [47, 55, 398, 421]
    TRAP_FIT_COEFF_BOT = [47, 55, 442, 377]

    MATLAB_TRAP_FIT_COEFF = [81, 125, 331, 350, -1448.2731999293467]
    MATLAB_BLINE_FIT_COEFF = [
        0.16890305013478935,
        -39.18378191113812,
        8256.387981752052,
    ]
    MATLAB_TRAP_FIT_COEFF_BOT = [80, 115, 382, 311, -1331.3180036312333]
    MATLAB_BLINE_FIT_COEFF_BOT = [
        0.20707462357475487,
        -44.91849057042588,
        8014.213383072912,
    ]

    SW_MATLAB = 4.972852917690252

    def setUp(self):
        self.slice_width = SliceWidth(
            input_data=[
                os.path.join(self.SLICE_WIDTH_DATA, "512_matrix", "512_matrix")
            ],
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )
        # self.file = str(TEST_DATA_DIR / 'slicewidth' / 'SLICEWIDTH' / '512_matrix')
        # self.dcm = dcmread(self.file)



"""Logger object and log decorator."""

import sys
import logging
import colorlog
from functools import wraps
from typing import Any, Callable, TypeVar


def configure_logger():
    """Configure logger for the standard out (command line) stream and save logs to file"""
    # make log formatters
    stream_formatter = colorlog.ColoredFormatter(
        "%(log_color)s%(asctime)-15s %(levelname).1s "
        "[%(filename)s:%(funcName)s:%(lineno)d] %(message)s",
        "%Y-%m-%d %H:%M:%S",
    )

    file_formatter = logging.Formatter(
        "%(asctime)-15s %(levelname).1s [%(filename)s:%(funcName)s:%(lineno)d]"
        " %(message)s",
        "%Y-%m-%d %H:%M:%S",
    )

    log_file = "Hazen_logger.log"
    stream_handler = logging.StreamHandler(sys.stdout)
    file_handler = logging.FileHandler(log_file)

    # set formatters
    stream_handler.setFormatter(stream_formatter)
    file_handler.setFormatter(file_formatter)

    # add handlers
    logger.addHandler(stream_handler)
    logger.addHandler(file_handler)


logger = logging.getLogger(__name__)
configure_logger()

T = TypeVar("T")


def log(fn: Callable[..., T]) -> Callable[..., T]:
    """Log the inputs and outputs to a function."""

    @wraps(fn)
    def inner(*args: Any, **kwargs: Any) -> T:
        logger.debug(
            "Calling %s with arguments: (%s) and keyword arguments: (%s)",
            fn.__name__,
            args,
            kwargs,
        )
        result = fn(*args, **kwargs)
        logger.debug("%s returned %s", fn.__name__, result)
        return result

    return inner



# Python imports
import logging
import sys
import tomllib
import unittest
from pathlib import Path

import hazenlib
from hazenlib.tasks.relaxometry import Relaxometry
from hazenlib.tasks.snr import SNR

from hazenlib.types import Measurement, Result
from hazenlib.utils import dcmread, get_dicom_files

# Local imports
from tests import TEST_DATA_DIR


logger = logging.getLogger(__name__)


class TestCliParser(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.file = str(
            TEST_DATA_DIR / "resolution" / "philips" / "IM-0004-0002.dcm"
        )
        self.dcm = dcmread(self.file)

    def test_version(self):
        pyproject_path = Path(__file__).parent.parent / "pyproject.toml"
        with pyproject_path.open("rb") as fp:
            pyproject_data = tomllib.load(fp)

        expected_version = pyproject_data["project"]["version"]
        self.assertIn(
            hazenlib.__version__,
            {expected_version, expected_version + "+dev"},
            (
                f"Package version {hazenlib.__version__} does not match"
                f" pyproject.toml version {expected_version}"
            ),
        )

    def test1_logger(self):
        path = str(TEST_DATA_DIR / "resolution" / "RESOLUTION")
        sys.argv = ["hazen", "spatial_resolution", path, "--log", "warning"]

        hazenlib.main()

        logging = hazenlib.logging

        self.assertEqual(logging.root.level, logging.WARNING)

    def test2_logger(self):
        path = str(TEST_DATA_DIR / "resolution" / "RESOLUTION")
        sys.argv = ["hazen", "spatial_resolution", path]

        hazenlib.main()

        logging = hazenlib.logging

        self.assertEqual(logging.root.level, logging.INFO)

    def test_snr_measured_slice_width(self):
        path = str(TEST_DATA_DIR / "snr" / "GE")
        files = get_dicom_files(path)
        snr_task = SNR(input_data=files, report=False, measured_slice_width=5)
        result = snr_task.run()

        measurements = [
            Measurement(
                name="SNR",
                subtype="subtraction",
                value=183.97,
                type="measured",
                visibility="intermediate",
            ),
            Measurement(
                name="SNR",
                subtype="subtraction",
                value=1518.61,
                type="normalised",
            ),
            Measurement(
                name="SNR",
                subtype="smoothing",
                value=184.41,
                type="measured",
                description="SNR_SAG_MEAS1_23_1",
                visibility="intermediate",
            ),
            Measurement(
                name="SNR",
                subtype="smoothing",
                value=1522.17,
                type="normalised",
                description="SNR_SAG_MEAS1_23_1",
            ),
            Measurement(
                name="SNR",
                subtype="smoothing",
                value=189.38,
                type="measured",
                description="SNR_SAG_MEAS2_24_1",
                visibility="intermediate",
            ),
            Measurement(
                name="SNR",
                subtype="smoothing",
                value=1563.2,
                type="normalised",
                description="SNR_SAG_MEAS2_24_1",
            ),
        ]

        dict1 = Result(
            task="SNR",
            files=["SNR_SAG_MEAS1_23_1", "SNR_SAG_MEAS2_24_1"],
        )

        for m in measurements:
            dict1.add_measurement(m)

        self.assertEqual(vars(result).keys(), vars(dict1).keys())
        for k, v in vars(result).items():
            # Metadata is dynamically populated and is tested elsewhere.
            if k != "measurement" and k != "metadata":
                self.assertEqual(v, vars(dict1)[k])

        for m_d in dict1.measurements:
            m_r = result.get_measurement(
                name=m_d.name,
                measurement_type=m_d.type,
                subtype=m_d.subtype,
                description=m_d.description,
                unit=m_d.unit,
            )[0]
            self.assertAlmostEqual(m_r.value, m_d.value)

    def test_relaxometry(self):
        path = str(
            TEST_DATA_DIR / "relaxometry" / "T1" / "site3_ge" / "plate4"
        )
        files = get_dicom_files(path)
        relaxometry_task = Relaxometry(input_data=files, report=False)
        result = relaxometry_task.run(calc="T1", plate_number=4, verbose=False)

        dict1 = Result(
            task="Relaxometry",
            files="Spin_Echo_34_2_4_t1",
        )

        measurement = Measurement(
            name="Relaxometry",
            type="measured",
            subtype="rms_frac_time_difference",
            value=0.135,
        )
        dict1.add_measurement(measurement)
        self.assertEqual(vars(dict1).keys(), vars(result).keys())
        self.assertAlmostEqual(
            measurement.value,
            result.get_measurement(name="Relaxometry")[0].value,
            4,
        )



import unittest
import pathlib

from hazenlib.utils import get_dicom_files
from hazenlib.tasks.acr_slice_position import ACRSlicePosition
from tests import TEST_DATA_DIR


class TestACRSlicePositionSiemens(unittest.TestCase):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens")
    slice_1_x_pts = [123, 129]
    slice_11_x_pts = [123, 129]
    slice_1_y_pts = [44, 84]
    slice_11_y_pts = [43, 83]
    dL = -0.59, -1.56

    def setUp(self):
        input_files = get_dicom_files(self.ACR_DATA)
        self.acr_slice_position_task = ACRSlicePosition(input_data=input_files)

        self.dcm_1 = self.acr_slice_position_task.ACR_obj.slice_stack[0]
        img_1 = self.dcm_1.pixel_array
        cxy_1, _ = self.acr_slice_position_task.ACR_obj.find_phantom_center(
            img_1,
            self.acr_slice_position_task.ACR_obj.dx,
            self.acr_slice_position_task.ACR_obj.dy,
        )
        self.slice1_x_pts, self.slice1_y_pts = (
            self.acr_slice_position_task.find_wedges(img_1, cxy_1)
        )

        self.dcm_11 = self.acr_slice_position_task.ACR_obj.slice_stack[-1]
        img_11 = self.dcm_11.pixel_array
        cxy_11, _ = self.acr_slice_position_task.ACR_obj.find_phantom_center(
            img_11,
            self.acr_slice_position_task.ACR_obj.dx,
            self.acr_slice_position_task.ACR_obj.dy,
        )
        (
            self.slice11_x_pts,
            self.slice11_y_pts,
        ) = self.acr_slice_position_task.find_wedges(img_11, cxy_11)

    # IMAGE 1
    def test_find_wedge_slice1_x(self):
        assert self.slice1_x_pts == self.slice_1_x_pts

    def test_find_wedge_slice1_y(self):
        assert self.slice1_y_pts == self.slice_1_y_pts

    # IMAGE 11
    def test_find_wedge_slice11_x(self):
        assert self.slice11_x_pts == self.slice_11_x_pts

    def test_find_wedge_slice11_y(self):
        assert self.slice11_y_pts == self.slice_11_y_pts

    def test_slice_position(self):
        slice_position_val_1 = round(
            self.acr_slice_position_task.get_slice_position(self.dcm_1), 2
        )
        slice_position_val_11 = round(
            self.acr_slice_position_task.get_slice_position(self.dcm_11), 2
        )

        print(
            "\ntest_slice_position.py::TestSlicePosition::test_slice_position"
        )
        print("new_release_value:", slice_position_val_1)
        print("fixed_value:", self.dL[0])

        assert slice_position_val_1 == self.dL[0]
        assert slice_position_val_11 == self.dL[1]


class TestACRSlicePositionGE(TestACRSlicePositionSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE")
    slice_1_x_pts = [247, 257]
    slice_11_x_pts = [246, 256]
    slice_1_y_pts = [93, 171]
    slice_11_y_pts = [94, 172]
    dL = 0.51, 4.47


class TestACRSlicePositionPhilipsAchieva(TestACRSlicePositionSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "PhilipsAchieva")
    slice_1_x_pts = [125, 131]
    slice_11_x_pts = [123, 129]
    slice_1_y_pts = [43, 83]
    slice_11_y_pts = [41, 81]
    dL = 1.37, -3.12


class TestACRSlicePositionSiemensSolaFit(TestACRSlicePositionSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "SiemensSolaFit")
    slice_1_x_pts = [124, 130]
    slice_11_x_pts = [123, 129]
    slice_1_y_pts = [43, 83]
    slice_11_y_pts = [41, 81]
    dL = -0.2, -1.37


class TestACRSlicePositionSiemensLargeSliceLocationDelta(
    TestACRSlicePositionSiemens
):
    ACR_DATA = pathlib.Path(
        TEST_DATA_DIR / "acr" / "SiemensLargeSliceLocationDelta"
    )
    slice_1_x_pts = [123, 129]
    slice_11_x_pts = [123, 129]
    slice_1_y_pts = [47, 87]
    slice_11_y_pts = [46, 86]
    dL = -6.64, -6.84


class TestACRSlicePositionPhilips3TDStream(TestACRSlicePositionSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Philips3TdStream")
    slice_1_x_pts = [124, 130]
    slice_11_x_pts = [124, 130]
    slice_1_y_pts = [44, 84]
    slice_11_y_pts = [41, 81]
    dL = 2.15, -2.15


class TestACRSlicePositionPhilips3TDStream2(TestACRSlicePositionSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Philips3TdStream2")
    slice_1_x_pts = [124, 130]
    slice_11_x_pts = [124, 130]
    slice_1_y_pts = [23, 63]
    slice_11_y_pts = [42, 82]
    dL = 3.12, -0.78


# triggers mispositioning of line profiles under a prior algorithm revision.
class TestACRSlicePositionSiemensSolaFit2(TestACRSlicePositionSiemens):
    ACR_DATA = pathlib.Path(
        TEST_DATA_DIR / "acr" / "SiemensSolaFitSlicePosition"
    )
    slice_1_x_pts = [124, 130]
    slice_11_x_pts = [124, 130]
    slice_1_y_pts = [43, 83]
    slice_11_y_pts = [42, 82]
    dL = -1.17, -2.54



"""Single point place to retrieve the `hazenlib` version.

This module provides a single canonical location for determining the
version string exposed as ``hazenlib.__version__``.

Version resolution strategy
---------------------------

* **Package vs distribution names**
  The importable package is named :mod:`hazenlib`, while the installed
  distribution (the project published on PyPI) is typically named
  ``hazen``. Because a single package can belong to one or more
  distributions, we first ask :func:`importlib.metadata.packages_distributions`
  which distribution name(s) provide the ``hazenlib`` package, and then
  query each distribution for its version.

* **Fallback to ``pyproject.toml``**
  When ``hazenlib`` is imported from a source tree without an installed
  distribution (for example, during development, in a local checkout, or
  in some test environments), querying the distribution metadata can
  fail. In that case, we fall back to reading the version declared in
  the top-level ``pyproject.toml`` under ``[project].version``.

* **``+dev`` suffix**
  If we have to fall back to ``pyproject.toml``, we append a ``\"+dev\"``
  suffix to the version. This indicates that the code is being run from
  a development (source) environment rather than from an installed,
  released distribution. This also avoids accidentally reporting an
  identical version string for both an installed release and a
  development checkout.

* **Order of precedence**
  1. Use the version from the first distribution associated with the
     ``hazenlib`` package whose metadata can be successfully read via
     :func:`importlib.metadata.version`.
  2. If no such distribution is found or all lookups fail, read the
     version from ``pyproject.toml`` and append ``\"+dev\"``.
"""

import importlib.metadata
import pathlib
import tomllib

distributions = importlib.metadata.packages_distributions().get("hazenlib", [])

__version__ = None
for dist_name in distributions:
    try:
        __version__ = importlib.metadata.version(dist_name)
        break
    except importlib.metadata.PackageNotFoundError:
        continue

if __version__ is None:
    pyproject_path = pathlib.Path(__file__).parent.parent / "pyproject.toml"
    try:
        with pyproject_path.open("rb") as f:
            data = tomllib.load(f)
    except tomllib.TOMLDecodeError as exc:
        raise RuntimeError(
            f"Failed to parse {pyproject_path} as TOML while resolving hazenlib version"
        ) from exc
    __version__ = data["project"]["version"] + "+dev"



"""Base class for Hazen tasks."""

# Typing imports
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pydicom

# Python imports
from pathlib import Path

# Module imports
from pydicom import dcmread

# Local imports
from hazenlib.logger import logger
from hazenlib.types import Result
from hazenlib.utils import REGEX_SCRUBNAME, scrub


class HazenTask:
    """Base class for performing tasks on image sets."""

    def __init__(
        self,
        input_data: Sequence,
        *,
        verbose: bool = False,
        report: bool = False,
        report_dir: str | Path | None = None,
    ) -> None:
        """Initialise a HazenTask instance.

        Args:
            input_data : list of filepaths to DICOM images
            report : Whether to create measurement visualisation diagrams.
                     Defaults to False.
            report_dir : Path to output report images. Defaults to None.

        """
        self._data_paths = [Path(p) for p in sorted(input_data)]
        self.dcm_list = [dcmread(dicom) for dicom in self._data_paths]

        # Log acquisition information for each DICOM file
        for dcm in self.dcm_list:
            acq_number = dcm.get("AcquisitionNumber", "N/A")
            series_desc = dcm.get("SeriesDescription", "N/A")
            series_number = dcm.get("SeriesNumber", "N/A")
            instance_number = dcm.get("InstanceNumber", "N/A")
            logger.info(
                "Loaded DICOM - AcquisitionNumber: %s, "
                "SeriesDescription: %s, SeriesNumber: %s, "
                "InstanceNumber: %s",
                acq_number,
                series_desc,
                series_number,
                instance_number,
            )

        self.report: bool = report
        self.report_path = (
            Path().cwd() / "report_image" / type(self).__name__
            if report_dir is None
            else Path(report_dir) / type(self).__name__
        )

        self.report_path.mkdir(parents=True, exist_ok=True)
        self.report_files: Sequence[str] = []

    def init_result_dict(self, desc: str = "") -> Result:
        """Initialise measurement results holder and input description.

        Returns
            d : holds measurement results and task input description.

        """
        return Result(
            task=type(self).__name__,
            desc=desc,
            files=[p.as_posix() for p in self._data_paths],
        )

    def img_desc(
        self,
        dcm: pydicom.Dataset,
        properties: Sequence | None = None,
    ) -> str:
        """Obtain values from the DICOM header to identify input series.

        Args:
            dcm : DICOM image object
            properties : list of DICOM header field names supported by pydicom
                that shuld be used to generate sereis identifier.
                Defaults to None.

        Returns:
            str: contatenation of the specified DICOM header property values

        """
        if properties is None:
            properties = [
                "SeriesDescription",
                "SeriesNumber",
                "InstanceNumber",
            ]
        try:
            metadata = [str(dcm.get(field)) for field in properties]
        except KeyError:
            logger.warning(
                f"Could not find one or more of the following "
                f"properties: {properties}",
            )
            metadata = [
                str(dcm.get(field))
                for field in ["SeriesDescription", "SeriesNumber"]
            ]

        join_char = "_"
        img_desc = join_char.join(metadata).replace(" ", join_char)
        # Let's make sure dirty names do not contaminate the file names
        # or any other string operations.
        return (
            scrub(
                img_desc,
                REGEX_SCRUBNAME,
                join_char,
            )
            .strip(join_char)
            .replace(join_char * 2, join_char)
        )
